import pytest
from uuid import UUID, uuid4
from typing import Any, Optional

from vellum.workflows.inputs import BaseInputs
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.nodes.core.inline_subworkflow_node.node import InlineSubworkflowNode
from vellum.workflows.nodes.core.retry_node.node import RetryNode
from vellum.workflows.nodes.core.templating_node.node import TemplatingNode
from vellum.workflows.nodes.core.try_node.node import TryNode
from vellum.workflows.nodes.displayable.final_output_node.node import FinalOutputNode
from vellum.workflows.references.lazy import LazyReference
from vellum.workflows.state.base import BaseState
from vellum.workflows.triggers.integration import IntegrationTrigger
from vellum.workflows.types.core import JsonObject
from vellum.workflows.workflows.base import BaseWorkflow
from vellum_ee.workflows.display.editor.types import NodeDisplayData, NodeDisplayPosition
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay
from vellum_ee.workflows.display.nodes.vellum.retry_node import BaseRetryNodeDisplay
from vellum_ee.workflows.display.nodes.vellum.try_node import BaseTryNodeDisplay
from vellum_ee.workflows.display.types import WorkflowDisplayContext
from vellum_ee.workflows.display.utils.exceptions import UserFacingException
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display


def test_serialize_workflow__node_referenced_in_workflow_outputs_not_in_graph():
    # GIVEN a couple of nodes
    class InNode(BaseNode):
        pass

    class OutNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    # AND A workflow that references the OutNode in its outputs but only has the InNode in its graph
    class Workflow(BaseWorkflow):
        graph = InNode

        class Outputs(BaseWorkflow.Outputs):
            final = OutNode.Outputs.foo

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=Workflow)

    # THEN it should raise an error
    with pytest.raises(UserFacingException) as exc_info:
        workflow_display.serialize()

    # AND the error message should be user friendly
    assert str(exc_info.value) == "Failed to serialize output 'final': Reference to node 'OutNode' not found in graph."


def test_serialize_workflow__workflow_outputs_reference_non_node_outputs():
    # GIVEN one Workflow
    class FirstWorkflow(BaseWorkflow):
        class Outputs(BaseWorkflow.Outputs):
            foo = "bar"

    # AND A workflow that references the Outputs of that Workflow
    class Workflow(BaseWorkflow):
        class Outputs(BaseWorkflow.Outputs):
            final = FirstWorkflow.Outputs.foo

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=Workflow)
    serialized_workflow = workflow_display.serialize()

    # THEN it should successfully serialize the workflow output reference to a constant
    assert isinstance(serialized_workflow, dict)
    output_variables = serialized_workflow["output_variables"]
    assert isinstance(output_variables, list)
    assert output_variables == [{"id": "2b32416b-ccfc-4231-a3a6-d08e76327815", "key": "final", "type": "STRING"}]

    # AND the output value should be a constant value
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    assert isinstance(workflow_raw_data, dict)
    output_values = workflow_raw_data["output_values"]
    assert isinstance(output_values, list)
    assert output_values == [
        {
            "output_variable_id": "2b32416b-ccfc-4231-a3a6-d08e76327815",
            "value": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "bar"}},
        }
    ]

    first_output_variable = output_variables[0]
    assert isinstance(first_output_variable, dict)
    first_output_value = output_values[0]
    assert isinstance(first_output_value, dict)
    assert first_output_variable["id"] == first_output_value["output_variable_id"]


def test_serialize_workflow__node_display_class_not_registered():
    # GIVEN a workflow with a node that has a display class referencing display data
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            result: str

    class StartNodeDisplay(BaseNodeDisplay[StartNode]):
        node_input_ids_by_name = {}
        display_data = NodeDisplayData(position=NodeDisplayPosition(x=0, y=0), width=None, height=None)

    class MyWorkflow(BaseWorkflow):
        graph = StartNode

        class Outputs(BaseWorkflow.Outputs):
            answer = StartNode.Outputs.result

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=MyWorkflow)
    data = workflow_display.serialize()

    # THEN it should should succeed
    assert data is not None


def test_get_event_display_context__node_display_filled_without_base_display():
    # GIVEN a simple workflow
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    class MyWorkflow(BaseWorkflow):
        graph = StartNode

    # WHEN we gather the event display context
    display_context = get_workflow_display(workflow_class=MyWorkflow).get_event_display_context()

    # THEN the node display should be included
    assert StartNode.__id__ in display_context.node_displays
    node_event_display = display_context.node_displays[StartNode.__id__]

    # AND so should their output ids
    assert StartNode.__output_ids__ == node_event_display.output_display


def test_get_event_display_context__node_display_filled_without_output_display():
    # GIVEN a simple workflow
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    class MyWorkflow(BaseWorkflow):
        graph = StartNode

    class StartNodeDisplay(BaseNodeDisplay[StartNode]):
        pass

    # WHEN we gather the event display context
    display_context = get_workflow_display(workflow_class=MyWorkflow).get_event_display_context()

    # THEN the node display should be included
    assert StartNode.__id__ in display_context.node_displays
    node_event_display = display_context.node_displays[StartNode.__id__]

    # AND so should their output ids
    assert node_event_display.output_display.keys() == {"foo"}


def test_get_event_display_context__node_display_to_include_subworkflow_display():
    # GIVEN a simple workflow
    class InnerNode(BaseNode):
        pass

    class Subworkflow(BaseWorkflow):
        graph = InnerNode

    # AND a workflow that includes the subworkflow
    class SubworkflowNode(InlineSubworkflowNode):
        subworkflow = Subworkflow

    class MyWorkflow(BaseWorkflow):
        graph = SubworkflowNode

    # WHEN we gather the event display context
    display_context = get_workflow_display(workflow_class=MyWorkflow).get_event_display_context()

    # THEN the subworkflow display should be included
    assert SubworkflowNode.__id__ in display_context.node_displays
    node_event_display = display_context.node_displays[SubworkflowNode.__id__]

    assert node_event_display.subworkflow_display is not None
    assert InnerNode.__id__ in node_event_display.subworkflow_display.node_displays


@pytest.mark.parametrize(
    ["AdornmentNode", "AdornmentNodeDisplay", "expected_adornment_output_names"],
    [
        [RetryNode, BaseRetryNodeDisplay, {"foo"}],
        [TryNode, BaseTryNodeDisplay, {"foo", "error"}],
    ],
    ids=["retry_node", "try_node"],
)
def test_get_event_display_context__node_display_for_adornment_nodes(
    AdornmentNode,
    AdornmentNodeDisplay,
    expected_adornment_output_names,
):
    # GIVEN a simple workflow with an adornment
    @AdornmentNode.wrap()
    class MyNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    class MyWorkflow(BaseWorkflow):
        graph = MyNode

    # AND a display class for the node
    adornment_node_id = uuid4()
    inner_node_id = uuid4()

    @AdornmentNodeDisplay.wrap(node_id=adornment_node_id)
    class MyNodeDisplay(BaseNodeDisplay[MyNode]):
        node_id = inner_node_id

    # WHEN we gather the event display context
    display_context = get_workflow_display(workflow_class=MyWorkflow).get_event_display_context()

    # THEN the subworkflow display should be included
    assert adornment_node_id in display_context.node_displays
    node_event_display = display_context.node_displays[adornment_node_id]
    assert node_event_display.subworkflow_display is not None
    assert inner_node_id in node_event_display.subworkflow_display.node_displays

    # AND the inner node should have the correct outputs
    inner_node_display = node_event_display.subworkflow_display.node_displays[inner_node_id]
    assert inner_node_display.output_display.keys() == {"foo"}
    assert node_event_display.output_display.keys() == expected_adornment_output_names


def test_get_event_display_context__templating_node_input_display():
    # GIVEN a simple workflow with a templating node referencing another node output
    class DataNode(BaseNode):
        class Outputs:
            bar: str

    class MyNode(TemplatingNode):
        inputs = {"foo": DataNode.Outputs.bar}

    class MyWorkflow(BaseWorkflow):
        graph = DataNode >> MyNode

    # WHEN we gather the event display context
    display_context = get_workflow_display(workflow_class=MyWorkflow).get_event_display_context()

    # THEN the subworkflow display should be included
    assert MyNode.__id__ in display_context.node_displays
    node_event_display = display_context.node_displays[MyNode.__id__]

    assert node_event_display.input_display.keys() == {"inputs.foo"}


def test_get_event_display_context__node_display_for_mutiple_adornments():
    # GIVEN a simple workflow with multiple adornments
    @TryNode.wrap()
    @RetryNode.wrap()
    class MyNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    class MyWorkflow(BaseWorkflow):
        graph = MyNode

    # AND a display class for the node
    node_id = uuid4()
    inner_node_id = uuid4()
    innermost_node_id = uuid4()

    @BaseTryNodeDisplay.wrap(node_id=node_id)
    @BaseRetryNodeDisplay.wrap(node_id=inner_node_id)
    class MyNodeDisplay(BaseNodeDisplay[MyNode]):
        node_id = innermost_node_id

    # WHEN we gather the event display context
    display_context = get_workflow_display(workflow_class=MyWorkflow).get_event_display_context()

    # THEN the subworkflow display should be included
    assert node_id in display_context.node_displays
    node_event_display = display_context.node_displays[node_id]
    assert node_event_display.subworkflow_display

    # AND the inner node should be included
    assert inner_node_id in node_event_display.subworkflow_display.node_displays
    inner_node_event_display = node_event_display.subworkflow_display.node_displays[inner_node_id]
    assert inner_node_event_display.subworkflow_display

    # AND the innermost node should be included
    assert innermost_node_id in inner_node_event_display.subworkflow_display.node_displays
    innermost_node_event_display = inner_node_event_display.subworkflow_display.node_displays[innermost_node_id]
    assert not innermost_node_event_display.subworkflow_display


def test_get_event_display_context__workflow_output_display_with_none():
    # GIVEN a workflow with a workflow output that is None
    class MyWorkflow(BaseWorkflow):
        class Outputs(BaseWorkflow.Outputs):
            foo = None
            bar = "baz"

    # WHEN we gather the event display context
    display_context = get_workflow_display(workflow_class=MyWorkflow).get_event_display_context()

    # THEN the workflow output display should be included
    assert display_context.workflow_outputs.keys() == {"foo", "bar"}


def test_get_event_display_context__trigger_attributes_included():
    """Trigger attributes should be included in workflow_inputs for display in executions list."""

    # GIVEN a workflow with an integration trigger that has attributes
    class SlackTrigger(IntegrationTrigger):
        message: str
        channel: str

        class Config:
            provider = "COMPOSIO"
            integration_name = "SLACK"
            slug = "slack_new_message"

    class ProcessNode(BaseNode):
        pass

    class MyWorkflow(BaseWorkflow):
        graph = SlackTrigger >> ProcessNode

    # WHEN we gather the event display context
    display_context = get_workflow_display(workflow_class=MyWorkflow).get_event_display_context()

    # THEN the trigger attributes should be included in workflow_inputs
    assert "message" in display_context.workflow_inputs
    assert "channel" in display_context.workflow_inputs

    # AND the IDs should match the trigger attribute IDs
    trigger_attr_refs = SlackTrigger.attribute_references()
    assert display_context.workflow_inputs["message"] == trigger_attr_refs["message"].id
    assert display_context.workflow_inputs["channel"] == trigger_attr_refs["channel"].id


def test_serialize_workflow__inherited_node_display_class_not_registered():
    # GIVEN a node meant to be used as a base
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            result: str

    # AND a node that inherits from it
    class InheritedNode(StartNode):
        foo: str

    # AND a workflow that uses the inherited node
    class MyWorkflow(BaseWorkflow):
        graph = InheritedNode

        class Outputs(BaseWorkflow.Outputs):
            answer = InheritedNode.Outputs.result

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=MyWorkflow)
    data = workflow_display.serialize()

    # THEN it should should succeed
    assert data is not None


def test_serialize_workflow__inherited_workflow_display_class_not_registered():
    # GIVEN a node
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            result: str

    # AND a workflow that uses the node
    class MyWorkflow(BaseWorkflow):
        graph = StartNode

    # AND a workflow that inherits from it
    class InheritedWorkflow(MyWorkflow):
        class Outputs(MyWorkflow.Outputs):
            answer = StartNode.Outputs.result

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=InheritedWorkflow)
    data = workflow_display.serialize()

    # THEN it should should succeed
    assert data is not None


def test_serialize_workflow__terminal_node_mismatches_workflow_output_name():
    # GIVEN a node
    class ExitNode(FinalOutputNode):
        class Outputs(FinalOutputNode.Outputs):
            value = "hello"

    # AND a workflow that uses the node
    class MyWorkflow(BaseWorkflow):
        graph = ExitNode

        class Outputs(BaseWorkflow.Outputs):
            answer = ExitNode.Outputs.value

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=MyWorkflow)
    data = workflow_display.serialize()

    # THEN it should have an output name that matches the workflow output
    assert isinstance(data["workflow_raw_data"], dict)
    assert isinstance(data["workflow_raw_data"]["nodes"], list)
    terminal_node = [
        node for node in data["workflow_raw_data"]["nodes"] if isinstance(node, dict) and node["type"] == "TERMINAL"
    ][0]
    assert isinstance(terminal_node["data"], dict)
    assert terminal_node["data"]["name"] == "answer"

    # AND the output variable should have the correct name
    assert isinstance(data["output_variables"], list)
    assert isinstance(data["output_variables"][0], dict)
    assert data["output_variables"][0]["key"] == "answer"
    assert data["output_variables"][0]["type"] == "STRING"

    # AND the output value should have the correct name
    output_variable_id = data["output_variables"][0]["id"]
    assert isinstance(data["workflow_raw_data"]["output_values"], list)
    assert isinstance(data["workflow_raw_data"]["output_values"][0], dict)
    assert data["workflow_raw_data"]["output_values"][0]["output_variable_id"] == output_variable_id
    assert data["workflow_raw_data"]["output_values"][0]["value"] == {
        "type": "NODE_OUTPUT",
        "node_id": str(ExitNode.__id__),
        "node_output_id": str(ExitNode.__output_ids__["value"]),
    }


def test_serialize_workflow__nested_lazy_reference():
    # GIVEN an inner node that references the output of an outer node
    class InnerNode(BaseNode):
        foo = LazyReference[str]("OuterNode.Outputs.bar")

        class Outputs(BaseNode.Outputs):
            foo = "foo"

    # AND a workflow that uses the inner node
    class InnerWorkflow(BaseWorkflow):
        graph = InnerNode

        class Outputs(BaseWorkflow.Outputs):
            foo = InnerNode.Outputs.foo

    # AND a subworkflow that uses the inner workflow
    class SubworkflowNode(InlineSubworkflowNode):
        subworkflow = InnerWorkflow

    # AND the outer node
    class OuterNode(BaseNode):

        class Outputs(BaseNode.Outputs):
            bar: str

    # AND a workflow that uses the subworkflow node and the outer node
    class MyWorkflow(BaseWorkflow):
        graph = SubworkflowNode >> OuterNode

        class Outputs(BaseWorkflow.Outputs):
            answer = SubworkflowNode.Outputs.foo

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=MyWorkflow)
    data: dict = workflow_display.serialize()

    # THEN it should have properly serialized the lazy reference
    subworkflow_node = next(
        node for node in data["workflow_raw_data"]["nodes"] if isinstance(node, dict) and node["type"] == "SUBWORKFLOW"
    )
    inner_node = next(
        node
        for node in subworkflow_node["data"]["workflow_raw_data"]["nodes"]
        if isinstance(node, dict) and node["type"] == "GENERIC"
    )

    assert inner_node["attributes"][0]["value"] == {
        "type": "NODE_OUTPUT",
        "node_id": str(OuterNode.__id__),
        "node_output_id": str(OuterNode.__output_ids__["bar"]),
    }


def test_serialize_workflow__array_values():
    # GIVEN a node with array and nested array values
    class MyNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            array_value = ["item1", "item2", "item3"]
            nested_array_value = [["item1", "item2", "item3"], ["item4", "item5", "item6"]]
            mixed_array_value = [["item1"], "item2", "item3"]

    # AND a workflow that uses these outputs
    class MyWorkflow(BaseWorkflow):
        graph = MyNode

        class Outputs(BaseWorkflow.Outputs):
            array_output = MyNode.Outputs.array_value
            nested_array_output = MyNode.Outputs.nested_array_value

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=MyWorkflow)
    data = workflow_display.serialize()

    # THEN it should properly serialize the array and dictionary values
    assert isinstance(data["workflow_raw_data"], dict)
    assert isinstance(data["workflow_raw_data"]["nodes"], list)
    raw_nodes = data["workflow_raw_data"]["nodes"]
    generic_nodes = [node for node in raw_nodes if isinstance(node, dict) and node["type"] == "GENERIC"]
    assert len(generic_nodes) > 0
    my_node = generic_nodes[0]

    outputs = my_node["outputs"]
    assert isinstance(outputs, list)

    array_outputs = [val for val in outputs if isinstance(val, dict) and val["name"] == "array_value"]
    assert len(array_outputs) > 0
    array_output = array_outputs[0]

    assert isinstance(array_output, dict)
    assert "value" in array_output
    assert array_output["value"] == {
        "type": "CONSTANT_VALUE",
        "value": {"type": "JSON", "value": ["item1", "item2", "item3"]},
    }

    nested_array_outputs = [val for val in outputs if isinstance(val, dict) and val["name"] == "nested_array_value"]
    assert len(nested_array_outputs) > 0
    nested_array_output = nested_array_outputs[0]

    assert isinstance(nested_array_output, dict)
    assert "value" in nested_array_output
    assert nested_array_output["value"] == {
        "type": "CONSTANT_VALUE",
        "value": {"type": "JSON", "value": [["item1", "item2", "item3"], ["item4", "item5", "item6"]]},
    }

    mixed_array_outputs = [val for val in outputs if isinstance(val, dict) and val["name"] == "mixed_array_value"]
    assert len(mixed_array_outputs) > 0
    mixed_array_output = mixed_array_outputs[0]

    assert isinstance(mixed_array_output, dict)
    assert "value" in mixed_array_output
    assert mixed_array_output["value"] == {
        "type": "CONSTANT_VALUE",
        "value": {"type": "JSON", "value": [["item1"], "item2", "item3"]},
    }


def test_serialize_workflow__array_reference():
    # GIVEN a node with array containing non-constant values (node references)
    class FirstNode(BaseNode):

        class Outputs(BaseNode.Outputs):
            value1: str
            value2: str

    class SecondNode(BaseNode):

        class Outputs(BaseNode.Outputs):
            # Array containing a mix of constants and node references
            mixed_array = ["constant1", FirstNode.Outputs.value1, "constant2", FirstNode.Outputs.value2]
            mixed_nested_array = [["constant1", FirstNode.Outputs.value1], ["constant2", FirstNode.Outputs.value2]]

    # AND a workflow that uses these outputs
    class MyWorkflow(BaseWorkflow):
        graph = FirstNode >> SecondNode

        class Outputs(BaseWorkflow.Outputs):
            mixed_array_output = SecondNode.Outputs.mixed_array
            mixed_nested_array_output = SecondNode.Outputs.mixed_nested_array

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=MyWorkflow)
    data = workflow_display.serialize()

    # THEN it should serialize as an ARRAY_REFERENCE
    assert isinstance(data["workflow_raw_data"], dict)
    assert isinstance(data["workflow_raw_data"]["nodes"], list)
    second_node = next(
        node
        for node in data["workflow_raw_data"]["nodes"]
        if isinstance(node, dict)
        and "definition" in node
        and isinstance(node["definition"], dict)
        and node["definition"]["name"] == "SecondNode"
    )
    assert isinstance(second_node, dict)

    assert "outputs" in second_node
    assert isinstance(second_node["outputs"], list)
    outputs = second_node["outputs"]

    mixed_array_outputs = [val for val in outputs if isinstance(val, dict) and val["name"] == "mixed_array"]
    assert len(mixed_array_outputs) > 0
    mixed_array_output = mixed_array_outputs[0]

    assert isinstance(mixed_array_output, dict)
    assert "value" in mixed_array_output
    assert mixed_array_output["value"] == {
        "type": "ARRAY_REFERENCE",
        "items": [
            {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "constant1"}},
            {
                "type": "NODE_OUTPUT",
                "node_id": "9a6037fd-e023-4331-8097-5144bacfc110",
                "node_output_id": "37521463-db12-41a3-ad6f-753165880356",
            },
            {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "constant2"}},
            {
                "type": "NODE_OUTPUT",
                "node_id": "9a6037fd-e023-4331-8097-5144bacfc110",
                "node_output_id": "b033bddf-987d-488d-8426-c5bb2dac7501",
            },
        ],
    }

    mixed_nested_array_outputs = [
        val for val in outputs if isinstance(val, dict) and val["name"] == "mixed_nested_array"
    ]
    assert len(mixed_nested_array_outputs) > 0
    mixed_nested_array_output = mixed_nested_array_outputs[0]

    assert isinstance(mixed_nested_array_output, dict)
    assert "value" in mixed_nested_array_output
    assert mixed_nested_array_output["value"] == {
        "type": "ARRAY_REFERENCE",
        "items": [
            {
                "type": "ARRAY_REFERENCE",
                "items": [
                    {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "constant1"}},
                    {
                        "type": "NODE_OUTPUT",
                        "node_id": "9a6037fd-e023-4331-8097-5144bacfc110",
                        "node_output_id": "37521463-db12-41a3-ad6f-753165880356",
                    },
                ],
            },
            {
                "type": "ARRAY_REFERENCE",
                "items": [
                    {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "constant2"}},
                    {
                        "type": "NODE_OUTPUT",
                        "node_id": "9a6037fd-e023-4331-8097-5144bacfc110",
                        "node_output_id": "b033bddf-987d-488d-8426-c5bb2dac7501",
                    },
                ],
            },
        ],
    }


def test_serialize_workflow__dict_values():
    # GIVEN a node with a dictionary value
    class MyNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            dict_value = {"key1": "value1", "key2": "value2"}
            nested_dict_value = {
                "key1": {"nested_key1": "value1", "nested_key2": "value2"},
                "key2": {"nested_key1": "value1", "nested_key2": "value2"},
            }
            mixed_dict_value = {"key1": "value1", "key2": {"key3": "value3", "key4": "value4"}}

    # AND a workflow that uses these outputs
    class MyWorkflow(BaseWorkflow):
        graph = MyNode

        class Outputs(BaseWorkflow.Outputs):
            dict_output = MyNode.Outputs.dict_value

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=MyWorkflow)
    data = workflow_display.serialize()

    # THEN it should serialize as a CONSTANT_VALUE
    assert isinstance(data["workflow_raw_data"], dict)
    assert isinstance(data["workflow_raw_data"]["nodes"], list)
    my_node = next(
        node for node in data["workflow_raw_data"]["nodes"] if isinstance(node, dict) and node["type"] == "GENERIC"
    )

    assert isinstance(my_node["outputs"], list)
    outputs = my_node["outputs"]

    dict_output = next(val for val in outputs if isinstance(val, dict) and val["name"] == "dict_value")
    assert isinstance(dict_output, dict)
    assert "value" in dict_output
    assert dict_output["value"] == {
        "type": "CONSTANT_VALUE",
        "value": {"type": "JSON", "value": {"key1": "value1", "key2": "value2"}},
    }

    nested_dict_output = next(val for val in outputs if isinstance(val, dict) and val["name"] == "nested_dict_value")
    assert isinstance(nested_dict_output, dict)
    assert "value" in nested_dict_output
    assert nested_dict_output["value"] == {
        "type": "CONSTANT_VALUE",
        "value": {
            "type": "JSON",
            "value": {
                "key1": {"nested_key1": "value1", "nested_key2": "value2"},
                "key2": {"nested_key1": "value1", "nested_key2": "value2"},
            },
        },
    }

    mixed_dict_output = next(val for val in outputs if isinstance(val, dict) and val["name"] == "mixed_dict_value")
    assert isinstance(mixed_dict_output, dict)
    assert "value" in mixed_dict_output
    assert mixed_dict_output["value"] == {
        "type": "CONSTANT_VALUE",
        "value": {"type": "JSON", "value": {"key1": "value1", "key2": {"key3": "value3", "key4": "value4"}}},
    }


def test_serialize_workflow__dict_reference():
    # GIVEN a node with a dictionary containing non-constant values (node references)
    class FirstNode(BaseNode):

        class Outputs(BaseNode.Outputs):
            value1: str

    class SecondNode(BaseNode):

        class Outputs(BaseNode.Outputs):
            # Dictionary containing a mix of constants and node references
            mixed_dict = {
                "key1": "constant1",
                "key2": FirstNode.Outputs.value1,
                "key3": "constant2",
                "key4": FirstNode.Outputs.value1,
            }
            mixed_nested_dict = {
                "key1": {"key1": "constant1", "key2": FirstNode.Outputs.value1},
                "key2": {"key1": "constant2", "key2": FirstNode.Outputs.value1},
            }

    # AND a workflow that uses these outputs
    class MyWorkflow(BaseWorkflow):
        graph = FirstNode >> SecondNode

        class Outputs(BaseWorkflow.Outputs):
            mixed_dict_output = SecondNode.Outputs.mixed_dict
            mixed_nested_dict_output = SecondNode.Outputs.mixed_nested_dict

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=MyWorkflow)
    data = workflow_display.serialize()

    # THEN it should serialize as a CONSTANT_VALUE
    assert isinstance(data["workflow_raw_data"], dict)
    assert isinstance(data["workflow_raw_data"]["nodes"], list)
    second_node = data["workflow_raw_data"]["nodes"][2]

    assert isinstance(second_node, dict)
    assert "outputs" in second_node
    assert isinstance(second_node["outputs"], list)

    outputs = second_node["outputs"]
    mixed_dict_output = next(val for val in outputs if isinstance(val, dict) and val["name"] == "mixed_dict")
    assert isinstance(mixed_dict_output, dict)
    assert "value" in mixed_dict_output
    assert mixed_dict_output["value"] == {
        "type": "DICTIONARY_REFERENCE",
        "entries": [
            {
                "id": "c717464d-127e-4970-a3d8-31761dd83deb",
                "key": "key1",
                "value": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "constant1"}},
            },
            {
                "id": "ca19400f-9cce-4b4f-860f-63c98f43ad88",
                "key": "key2",
                "value": {
                    "type": "NODE_OUTPUT",
                    "node_id": "0f81f7b9-392b-4f0e-8584-0ff040fba961",
                    "node_output_id": "0b63e869-e978-4ec9-9f47-0cc1c7e22076",
                },
            },
            {
                "id": "10df74b1-56f8-4414-8042-3d4f5a51314d",
                "key": "key3",
                "value": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "constant2"}},
            },
            {
                "id": "4803c0f8-e582-4693-b9c3-1d6d54f999b1",
                "key": "key4",
                "value": {
                    "type": "NODE_OUTPUT",
                    "node_id": "0f81f7b9-392b-4f0e-8584-0ff040fba961",
                    "node_output_id": "0b63e869-e978-4ec9-9f47-0cc1c7e22076",
                },
            },
        ],
    }

    mixed_nested_dict_output = next(
        val for val in outputs if isinstance(val, dict) and val["name"] == "mixed_nested_dict"
    )
    assert isinstance(mixed_nested_dict_output, dict)
    assert "value" in mixed_nested_dict_output
    assert mixed_nested_dict_output["value"] == {
        "type": "DICTIONARY_REFERENCE",
        "entries": [
            {
                "id": "c717464d-127e-4970-a3d8-31761dd83deb",
                "key": "key1",
                "value": {
                    "type": "DICTIONARY_REFERENCE",
                    "entries": [
                        {
                            "id": "c717464d-127e-4970-a3d8-31761dd83deb",
                            "key": "key1",
                            "value": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "constant1"}},
                        },
                        {
                            "id": "ca19400f-9cce-4b4f-860f-63c98f43ad88",
                            "key": "key2",
                            "value": {
                                "type": "NODE_OUTPUT",
                                "node_id": "0f81f7b9-392b-4f0e-8584-0ff040fba961",
                                "node_output_id": "0b63e869-e978-4ec9-9f47-0cc1c7e22076",
                            },
                        },
                    ],
                },
            },
            {
                "id": "ca19400f-9cce-4b4f-860f-63c98f43ad88",
                "key": "key2",
                "value": {
                    "type": "DICTIONARY_REFERENCE",
                    "entries": [
                        {
                            "id": "c717464d-127e-4970-a3d8-31761dd83deb",
                            "key": "key1",
                            "value": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "constant2"}},
                        },
                        {
                            "id": "ca19400f-9cce-4b4f-860f-63c98f43ad88",
                            "key": "key2",
                            "value": {
                                "type": "NODE_OUTPUT",
                                "node_id": "0f81f7b9-392b-4f0e-8584-0ff040fba961",
                                "node_output_id": "0b63e869-e978-4ec9-9f47-0cc1c7e22076",
                            },
                        },
                    ],
                },
            },
        ],
    }


def test_serialize_workflow__empty_rules_indexerror():
    """Test that workflow serialization handles dictionary key access correctly."""

    # GIVEN a node with dictionary output
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            data: dict = {"key": "value"}

    # AND a workflow that references dictionary key access in its outputs
    class MyWorkflow(BaseWorkflow):
        graph = StartNode

        class Outputs(BaseWorkflow.Outputs):
            # This dictionary key access should be handled gracefully
            problematic_output = StartNode.Outputs.data["bar"]

    # WHEN we serialize the workflow
    workflow_display = get_workflow_display(workflow_class=MyWorkflow)
    result: dict = workflow_display.serialize()

    assert result is not None
    assert "output_variables" in result
    assert "workflow_raw_data" in result

    # AND the workflow output should contain the dictionary key access
    output_variables = result["output_variables"]
    assert len(output_variables) == 1
    assert output_variables[0]["key"] == "problematic_output"


def test_serialize_workflow__input_variables():
    # GIVEN a workflow with inputs
    class Inputs(BaseInputs):
        empty_string: str = ""
        input_1: str
        input_2: Optional[str]
        input_3: int = 1
        input_4: Optional[int] = 2

    class MyWorkflow(BaseWorkflow[Inputs, BaseState]):
        pass

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=MyWorkflow)
    data = workflow_display.serialize()

    # THEN the inputs should be serialized correctly
    assert "input_variables" in data
    input_variables = data["input_variables"]
    assert isinstance(input_variables, list)
    assert len(input_variables) == 5

    empty_string = next(var for var in input_variables if isinstance(var, dict) and var["key"] == "empty_string")
    assert empty_string == {
        "id": "09c3e825-8932-4d37-990a-b8a7c2a3bdec",
        "key": "empty_string",
        "type": "STRING",
        "default": {"type": "STRING", "value": ""},
        "required": False,
        "extensions": {"color": None},
        "schema": {"type": "string"},
    }

    input_1 = next(var for var in input_variables if isinstance(var, dict) and var["key"] == "input_1")
    assert input_1 == {
        "id": "13bd7980-3fbd-486c-9ebd-a29d84f7bda0",
        "key": "input_1",
        "type": "STRING",
        "default": None,
        "required": True,
        "extensions": {"color": None},
        "schema": {"type": "string"},
    }

    input_2 = next(var for var in input_variables if isinstance(var, dict) and var["key"] == "input_2")
    assert input_2 == {
        "id": "13847952-beab-408d-945e-cfa079e6e124",
        "key": "input_2",
        "type": "STRING",
        "default": None,
        "required": False,
        "extensions": {"color": None},
        "schema": {"anyOf": [{"type": "string"}, {"type": "null"}]},
    }

    input_3 = next(var for var in input_variables if isinstance(var, dict) and var["key"] == "input_3")
    assert input_3 == {
        "id": "2e38e1a4-09ff-4bb8-a12e-9bf54d4f3a5e",
        "key": "input_3",
        "type": "NUMBER",
        "default": {"type": "NUMBER", "value": 1.0},
        "required": False,
        "extensions": {"color": None},
        "schema": {"type": "integer"},
    }

    input_4 = next(var for var in input_variables if isinstance(var, dict) and var["key"] == "input_4")
    assert input_4 == {
        "id": "d945b6ae-2490-4bfb-9b1c-b1e484dfd4f6",
        "key": "input_4",
        "type": "NUMBER",
        "default": {"type": "NUMBER", "value": 2.0},
        "required": False,
        "extensions": {"color": None},
        "schema": {"anyOf": [{"type": "integer"}, {"type": "null"}]},
    }


def test_serialize_workflow__state_variables():
    # GIVEN a workflow with state variables
    class State(BaseState):
        empty_string: str = ""
        state_1: str = "hello"
        state_2: Optional[str] = None
        state_3: int = 1
        state_4: Optional[int] = 2

    class MyWorkflow(BaseWorkflow[BaseInputs, State]):
        pass

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=MyWorkflow)
    data = workflow_display.serialize()

    # THEN the state variables should be serialized correctly
    assert "state_variables" in data
    state_variables = data["state_variables"]
    assert isinstance(state_variables, list)
    assert len(state_variables) == 5

    empty_string = next(var for var in state_variables if isinstance(var, dict) and var["key"] == "empty_string")
    assert empty_string == {
        "id": "c69e2507-f610-4a6f-84cc-a5bc2aa48551",
        "key": "empty_string",
        "type": "STRING",
        "default": {"type": "STRING", "value": ""},
        "required": False,
        "extensions": {"color": None},
    }

    state_1 = next(var for var in state_variables if isinstance(var, dict) and var["key"] == "state_1")
    assert state_1 == {
        "id": "151113d2-9bbf-428d-a1c1-0a9cf4fdedf3",
        "key": "state_1",
        "type": "STRING",
        "default": {"type": "STRING", "value": "hello"},
        "required": False,
        "extensions": {"color": None},
    }

    state_2 = next(var for var in state_variables if isinstance(var, dict) and var["key"] == "state_2")
    assert state_2 == {
        "id": "9a8d7a55-8bd2-497d-820c-dee665144a48",
        "key": "state_2",
        "type": "STRING",
        "default": None,
        "required": False,
        "extensions": {"color": None},
    }

    state_3 = next(var for var in state_variables if isinstance(var, dict) and var["key"] == "state_3")
    assert state_3 == {
        "id": "ffde4327-12c4-4c55-82d6-3ab88f0b1037",
        "key": "state_3",
        "type": "NUMBER",
        "default": {"type": "NUMBER", "value": 1.0},
        "required": False,
        "extensions": {"color": None},
    }

    state_4 = next(var for var in state_variables if isinstance(var, dict) and var["key"] == "state_4")
    assert state_4 == {
        "id": "2467c1e6-b6aa-42d7-b079-84c8a650fbca",
        "key": "state_4",
        "type": "NUMBER",
        "default": {"type": "NUMBER", "value": 2.0},
        "required": False,
        "extensions": {"color": None},
    }


def test_serialize_workflow__with_complete_node_failure_prunes_edges():
    """Test that edges are pruned when a node completely fails to serialize (serialized_node is null)."""

    # GIVEN a node that completely fails to serialize
    class FailingNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            result: str

    class FailingNodeDisplay(BaseNodeDisplay[FailingNode]):
        def serialize(
            self, display_context: WorkflowDisplayContext, error_output_id: Optional[UUID] = None, **kwargs: Any
        ) -> JsonObject:
            raise NotImplementedError("Complete node serialization failure")

    # AND a workflow with the failing node connected to another node
    class WorkingNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            value: str

    class Workflow(BaseWorkflow):
        graph = FailingNode >> WorkingNode

    # WHEN we serialize the workflow with dry_run=True
    workflow_display = get_workflow_display(workflow_class=Workflow, dry_run=True)

    # AND we register the failing display class
    workflow_display.display_context.node_displays[FailingNode] = FailingNodeDisplay()

    data: dict = workflow_display.serialize()

    # THEN the workflow should serialize but with no edges (pruned due to invalid node)
    assert data["workflow_raw_data"]["edges"] == []

    # AND only the working node and entrypoint should be in the serialized nodes
    assert len(data["workflow_raw_data"]["nodes"]) == 2
    node_types = [node["type"] for node in data["workflow_raw_data"]["nodes"]]
    assert "ENTRYPOINT" in node_types
    assert "GENERIC" in node_types  # This is the WorkingNode that should still be serialized


def test_serialize_workflow__node_with_invalid_input_reference():
    """Test that serialization captures errors when nodes reference a non-existent input attribute."""

    # GIVEN a workflow with defined inputs
    class Inputs(BaseInputs):
        valid_input: str

    # AND a templating node that references a non-existent input
    class MyTemplatingNode(TemplatingNode):
        class Outputs(TemplatingNode.Outputs):
            pass

        template = "valid: {{ valid_input }}, invalid: {{ invalid_ref }}"
        inputs = {
            "valid_input": Inputs.valid_input,
            "invalid_ref": Inputs.invalid_ref,
        }

    # AND a base node that also references the non-existent input
    class MyBaseNode(BaseNode):
        invalid_ref = Inputs.invalid_ref

        class Outputs(BaseNode.Outputs):
            result: str

        def run(self) -> BaseNode.Outputs:
            return self.Outputs(result="done")

    class MyBaseNodeDisplay(BaseNodeDisplay[MyBaseNode]):
        __serializable_inputs__ = {MyBaseNode.invalid_ref}

    # WHEN we create a workflow with both nodes and serialize with dry_run=True
    class MyWorkflow(BaseWorkflow[Inputs, BaseState]):
        graph = MyTemplatingNode >> MyBaseNode

    workflow_display = get_workflow_display(workflow_class=MyWorkflow, dry_run=True)
    serialized = workflow_display.serialize()

    # THEN the serialization should succeed without raising an exception
    assert serialized is not None
    assert "workflow_raw_data" in serialized

    errors = list(workflow_display.display_context.errors)
    assert len(errors) > 0

    # AND the error messages should reference the missing attribute
    error_messages = [str(e) for e in errors]
    assert any("invalid_ref" in msg for msg in error_messages)

    invalid_nodes = list(workflow_display.display_context.invalid_nodes)
    assert len(invalid_nodes) >= 2



# This file was auto-generated by Fern from our API Definition.

import typing

from .node_execution_fulfilled_event import NodeExecutionFulfilledEvent
from .node_execution_rejected_event import NodeExecutionRejectedEvent

WorkflowSandboxExecuteNodeResponse = typing.Union[NodeExecutionFulfilledEvent, NodeExecutionRejectedEvent]



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .test_suite_run_prompt_sandbox_history_item_exec_config_data_request import (
    TestSuiteRunPromptSandboxHistoryItemExecConfigDataRequest,
)


class TestSuiteRunPromptSandboxHistoryItemExecConfigRequest(UniversalBaseModel):
    """
    Execution configuration for running a Test Suite against a Prompt Sandbox History Item
    """

    type: typing.Literal["PROMPT_SANDBOX_HISTORY_ITEM"] = "PROMPT_SANDBOX_HISTORY_ITEM"
    data: TestSuiteRunPromptSandboxHistoryItemExecConfigDataRequest
    test_case_ids: typing.Optional[typing.List[str]] = pydantic.Field(default=None)
    """
    Optionally specify a subset of test case ids to run. If not provided, all test cases within the test suite will be run by default.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_tool_calling_node_inline_workflow_tool_wrapper.workflow import (
    BasicToolCallingNodeInlineWorkflowToolWrapperWorkflow,
)


def test_serialize_workflow__inline_workflow_with_tool_wrapper():
    """
    Tests that an inline workflow with tool wrapper serializes correctly with definition containing inputs and examples.
    """

    # GIVEN a workflow that uses an inline workflow with tool wrapper
    workflow_class = BasicToolCallingNodeInlineWorkflowToolWrapperWorkflow

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=workflow_class)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should include both query and context
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 2
    input_keys = {var["key"] for var in input_variables}
    assert input_keys == {"query", "context"}

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 2
    output_keys = {var["key"] for var in output_variables}
    assert output_keys == {"text", "chat_history"}

    # AND the inline workflow tool should have the definition attribute serialized
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    tool_calling_node = workflow_raw_data["nodes"][1]
    function_attributes = next(
        attribute for attribute in tool_calling_node["attributes"] if attribute["name"] == "functions"
    )
    assert function_attributes["value"]["type"] == "CONSTANT_VALUE"
    assert function_attributes["value"]["value"]["type"] == "JSON"
    inline_workflow_tool = function_attributes["value"]["value"]["value"][0]

    # AND the inline workflow tool should have the correct type
    assert inline_workflow_tool["type"] == "INLINE_WORKFLOW"
    assert inline_workflow_tool["name"] == "BasicInlineSubworkflowWorkflow"

    # AND the inline workflow tool should have a definition attribute (like code tool)
    assert "definition" in inline_workflow_tool
    definition = inline_workflow_tool["definition"]

    # AND the definition should match the expected structure with inputs and examples
    context_var = next(var for var in input_variables if var["key"] == "context")
    context_input_variable_id = context_var["id"]

    assert definition == {
        "state": None,
        "cache_config": None,
        "name": "basic_inline_subworkflow_workflow",
        "description": "\n    A workflow that gets the weather for a given city and date with context.\n    ",
        "parameters": {
            "type": "object",
            "properties": {
                "city": {"type": "string"},
                "date": {"type": "string"},
            },
            "required": ["city", "date"],
            "examples": [{"city": "San Francisco", "date": "2025-01-01"}],
        },
        "inputs": {
            "context": {
                "type": "WORKFLOW_INPUT",
                "input_variable_id": context_input_variable_id,
            }
        },
        "forced": None,
        "strict": None,
    }



"""Namespace utilities for workflow server."""

import random
import string


def get_random_namespace() -> str:
    """
    Generate a random namespace for virtual file loading.

    This generates a workflow_tmp_* namespace like vembda uses for loading
    workflows dynamically.

    Returns:
        A random namespace string in the format "workflow_tmp_{random_suffix}"
    """
    random_suffix = "".join(random.choices(string.ascii_letters + string.digits, k=16))
    return f"workflow_tmp_{random_suffix}"



import pytest
from unittest.mock import patch

from vellum.utils.vellum_client import create_vellum_client, create_vellum_environment


class TestCreateVellumClient:
    """Tests for create_vellum_client and create_vellum_environment."""

    @pytest.mark.parametrize(
        ["api_url", "predict_api_url", "expected_predict_url"],
        [
            (None, None, "https://predict.vellum.ai"),
            (None, "https://custom-predict.example.com", "https://custom-predict.example.com"),
            ("CUSTOM_API_URL", None, "https://predict.vellum.ai"),
            ("CUSTOM_API_URL", "https://custom-predict.example.com", "https://custom-predict.example.com"),
        ],
        ids=[
            "defaults_to_standard_predict_url",
            "predict_api_url_overrides_default",
            "api_url_env_var_does_not_affect_predict_when_not_set",
            "predict_api_url_takes_precedence_over_api_url",
        ],
    )
    def test_create_vellum_client__predict_url_resolution(self, api_url, predict_api_url, expected_predict_url):
        """Tests that predict_api_url is resolved correctly based on provided parameters."""

        # GIVEN no environment variables are set
        with patch.dict("os.environ", {}, clear=True):
            # WHEN creating a vellum client with the given parameters
            client = create_vellum_client(
                api_key="test-api-key",
                api_url=api_url,
                predict_api_url=predict_api_url,
            )

            # THEN the predict URL should be resolved correctly
            assert client._client_wrapper._environment.predict == expected_predict_url

    @pytest.mark.parametrize(
        ["env_vars", "predict_api_url", "expected_predict_url"],
        [
            ({"VELLUM_PREDICT_API_URL": "https://env-predict.example.com"}, None, "https://env-predict.example.com"),
            (
                {"VELLUM_PREDICT_API_URL": "https://env-predict.example.com"},
                "https://param-predict.example.com",
                "https://param-predict.example.com",
            ),
            ({"VELLUM_API_URL": "https://env-api.example.com"}, None, "https://env-api.example.com"),
            (
                {
                    "VELLUM_API_URL": "https://env-api.example.com",
                    "VELLUM_PREDICT_API_URL": "https://env-predict.example.com",
                },
                None,
                "https://env-predict.example.com",
            ),
        ],
        ids=[
            "env_var_predict_api_url_is_used",
            "param_predict_api_url_overrides_env_var",
            "env_var_api_url_fallback_for_predict",
            "env_var_predict_api_url_takes_precedence_over_api_url",
        ],
    )
    def test_create_vellum_client__predict_url_env_var_resolution(
        self, env_vars, predict_api_url, expected_predict_url
    ):
        """Tests that predict_api_url respects environment variables with correct precedence."""

        # GIVEN specific environment variables are set
        with patch.dict("os.environ", env_vars, clear=True):
            # WHEN creating a vellum client with the given parameters
            client = create_vellum_client(
                api_key="test-api-key",
                predict_api_url=predict_api_url,
            )

            # THEN the predict URL should be resolved correctly
            assert client._client_wrapper._environment.predict == expected_predict_url

    def test_create_vellum_environment__predict_api_url_does_not_affect_other_urls(self):
        """Tests that predict_api_url only affects the predict endpoint, not default or documents."""

        # GIVEN no environment variables are set
        with patch.dict("os.environ", {}, clear=True):
            # WHEN creating a vellum environment with a custom predict_api_url
            environment = create_vellum_environment(predict_api_url="https://custom-predict.example.com")

            # THEN the predict URL should be customized
            assert environment.predict == "https://custom-predict.example.com"

            # AND the default and documents URLs should remain unchanged
            assert environment.default == "https://api.vellum.ai"
            assert environment.documents == "https://documents.vellum.ai"



"""Tests for multiple triggers pointing to the same node."""

from vellum.workflows import BaseWorkflow
from vellum.workflows.constants import VellumIntegrationProviderType
from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.state.base import BaseState
from vellum.workflows.triggers.integration import IntegrationTrigger
from vellum.workflows.triggers.manual import ManualTrigger
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display


def test_manual_and_slack_trigger_same_node():
    """
    Tests that when Manual and Slack triggers point to the same node,
    we have edges for both triggers.
    """

    class MyManualTrigger(ManualTrigger):
        pass

    class SlackMessageTrigger(IntegrationTrigger):
        message: str
        channel: str

        class Config(IntegrationTrigger.Config):
            provider = VellumIntegrationProviderType.COMPOSIO
            integration_name = "SLACK"
            slug = "slack_new_message"

    class ProcessNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            result: str

        def run(self) -> Outputs:
            return self.Outputs(result="processed")

    class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = {
            MyManualTrigger >> ProcessNode,
            SlackMessageTrigger >> ProcessNode,
        }

    result = get_workflow_display(workflow_class=TestWorkflow).serialize()

    assert "triggers" in result
    triggers = result["triggers"]
    assert isinstance(triggers, list)
    assert len(triggers) == 2

    manual_trigger = next((t for t in triggers if isinstance(t, dict) and t.get("type") == "MANUAL"), None)
    assert manual_trigger is not None, "Should have a Manual trigger"
    assert isinstance(manual_trigger, dict)
    manual_trigger_id = manual_trigger["id"]

    slack_trigger = next((t for t in triggers if isinstance(t, dict) and t.get("type") == "INTEGRATION"), None)
    assert slack_trigger is not None, "Should have a Slack trigger"
    assert isinstance(slack_trigger, dict)
    slack_trigger_id = slack_trigger["id"]

    workflow_raw_data = result["workflow_raw_data"]
    assert isinstance(workflow_raw_data, dict)
    edges = workflow_raw_data["edges"]
    assert isinstance(edges, list)

    nodes = workflow_raw_data["nodes"]
    assert isinstance(nodes, list)
    process_nodes = [n for n in nodes if isinstance(n, dict) and n.get("type") not in ("TERMINAL", "ENTRYPOINT")]
    assert len(process_nodes) > 0, "Should have at least one process node"
    process_node = process_nodes[0]
    process_node_id = process_node["id"]

    manual_edge = next(
        (
            e
            for e in edges
            if isinstance(e, dict)
            and e.get("source_node_id") == manual_trigger_id
            and e.get("target_node_id") == process_node_id
        ),
        None,
    )
    assert manual_edge is not None, (
        f"Should have edge from Manual trigger ({manual_trigger_id}) " f"to ProcessNode ({process_node_id})"
    )

    slack_edge = next(
        (
            e
            for e in edges
            if isinstance(e, dict)
            and e.get("source_node_id") == slack_trigger_id
            and e.get("target_node_id") == process_node_id
        ),
        None,
    )
    assert slack_edge is not None, (
        f"Should have edge from Slack trigger ({slack_trigger_id}) " f"to ProcessNode ({process_node_id})"
    )


def test_two_integration_triggers_same_node_unique_edge_ids():
    """
    Tests that when two IntegrationTriggers point to the same node,
    each trigger edge gets a unique ID (not duplicates).
    """

    # GIVEN two different IntegrationTriggers
    class SlackMessageTrigger(IntegrationTrigger):
        message: str

        class Config(IntegrationTrigger.Config):
            provider = VellumIntegrationProviderType.COMPOSIO
            integration_name = "SLACK"
            slug = "slack_new_message"

    class GithubPRTrigger(IntegrationTrigger):
        pr_title: str

        class Config(IntegrationTrigger.Config):
            provider = VellumIntegrationProviderType.COMPOSIO
            integration_name = "GITHUB"
            slug = "github_pr_opened"

    class ProcessNode(BaseNode):
        pass

    # AND a workflow where both triggers point to the same node
    class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = {
            SlackMessageTrigger >> ProcessNode,
            GithubPRTrigger >> ProcessNode,
        }

    # WHEN we serialize the workflow
    result = get_workflow_display(workflow_class=TestWorkflow).serialize()

    # THEN we should have two triggers
    triggers = result["triggers"]
    assert isinstance(triggers, list)
    assert len(triggers) == 2

    def get_integration_name(trigger: dict) -> str:
        exec_config = trigger.get("exec_config")
        if isinstance(exec_config, dict):
            name = exec_config.get("integration_name")
            return str(name) if name else ""
        return ""

    slack_trigger = next(
        (t for t in triggers if isinstance(t, dict) and get_integration_name(t) == "SLACK"),
        None,
    )
    assert slack_trigger is not None
    assert isinstance(slack_trigger, dict)
    slack_trigger_id = slack_trigger["id"]

    github_trigger = next(
        (t for t in triggers if isinstance(t, dict) and get_integration_name(t) == "GITHUB"),
        None,
    )
    assert github_trigger is not None
    assert isinstance(github_trigger, dict)
    github_trigger_id = github_trigger["id"]

    # AND we should have edges from both triggers
    workflow_raw_data = result["workflow_raw_data"]
    assert isinstance(workflow_raw_data, dict)
    edges = workflow_raw_data["edges"]
    assert isinstance(edges, list)

    nodes = workflow_raw_data["nodes"]
    assert isinstance(nodes, list)
    process_nodes = [n for n in nodes if isinstance(n, dict) and n.get("type") not in ("TERMINAL", "ENTRYPOINT")]
    assert len(process_nodes) > 0
    process_node = process_nodes[0]
    assert isinstance(process_node, dict)
    process_node_id = process_node["id"]

    slack_edge = next(
        (
            e
            for e in edges
            if isinstance(e, dict)
            and e.get("source_node_id") == slack_trigger_id
            and e.get("target_node_id") == process_node_id
        ),
        None,
    )
    assert slack_edge is not None, "Should have edge from Slack trigger to ProcessNode"
    assert isinstance(slack_edge, dict)

    github_edge = next(
        (
            e
            for e in edges
            if isinstance(e, dict)
            and e.get("source_node_id") == github_trigger_id
            and e.get("target_node_id") == process_node_id
        ),
        None,
    )
    assert github_edge is not None, "Should have edge from GitHub trigger to ProcessNode"
    assert isinstance(github_edge, dict)

    # AND the edge IDs should be unique (not duplicates)
    assert slack_edge["id"] != github_edge["id"], (
        f"Edge IDs should be unique but both are {slack_edge['id']}. "
        "Multiple triggers targeting the same node should have distinct edge IDs."
    )



"""Tests for partial WorkflowMetaDisplay override serialization."""

from vellum.workflows import BaseWorkflow
from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.state.base import BaseState
from vellum_ee.workflows.display.base import WorkflowDisplayData, WorkflowDisplayDataViewport, WorkflowMetaDisplay
from vellum_ee.workflows.display.workflows.base_workflow_display import BaseWorkflowDisplay
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display


def test_triggerless_workflow_with_partial_display_override_creates_entrypoint():
    """
    Tests that a triggerless workflow with partial WorkflowMetaDisplay overrides
    (only display_data set) still creates an ENTRYPOINT node with backfilled IDs.
    """

    # GIVEN a simple triggerless workflow
    class ProcessNode(BaseNode):
        pass

    class TestWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = ProcessNode

    # AND a display class that only overrides display_data (viewport), not entrypoint IDs
    class TestWorkflowDisplay(BaseWorkflowDisplay[TestWorkflow]):
        workflow_display = WorkflowMetaDisplay(
            display_data=WorkflowDisplayData(viewport=WorkflowDisplayDataViewport(x=100.0, y=200.0, zoom=1.5))
        )

    # WHEN we serialize the workflow
    result = get_workflow_display(workflow_class=TestWorkflow).serialize()

    # THEN the workflow should have an ENTRYPOINT node
    workflow_raw_data = result["workflow_raw_data"]
    assert isinstance(workflow_raw_data, dict)
    nodes = workflow_raw_data["nodes"]
    assert isinstance(nodes, list)
    entrypoint_nodes = [n for n in nodes if isinstance(n, dict) and n.get("type") == "ENTRYPOINT"]
    assert len(entrypoint_nodes) == 1, "Triggerless workflow with partial overrides should have an ENTRYPOINT node"

    # AND the ENTRYPOINT node should have valid IDs (not None)
    entrypoint_node = entrypoint_nodes[0]
    assert isinstance(entrypoint_node, dict)
    assert entrypoint_node["id"] is not None, "ENTRYPOINT node should have a non-None id"
    entrypoint_data = entrypoint_node["data"]
    assert isinstance(entrypoint_data, dict)
    source_handle_id = entrypoint_data["source_handle_id"]
    assert source_handle_id is not None, "ENTRYPOINT node should have a non-None source_handle_id"

    # AND there should be an edge from ENTRYPOINT to the process node
    edges = workflow_raw_data["edges"]
    assert isinstance(edges, list)
    entrypoint_edges = [e for e in edges if isinstance(e, dict) and e.get("source_node_id") == entrypoint_node["id"]]
    assert len(entrypoint_edges) > 0, "Should have edges from ENTRYPOINT node"



from functools import cached_property
from uuid import UUID
from typing import TYPE_CHECKING, Optional, Tuple, Type, TypeVar, cast

from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException
from vellum.workflows.utils.uuids import get_state_value_id

if TYPE_CHECKING:
    from vellum.workflows.state.base import BaseState


_T = TypeVar("_T")


class StateValueReference(BaseDescriptor[_T]):

    def __init__(
        self,
        *,
        name: str,
        types: Tuple[Type[_T], ...],
        instance: Optional[_T],
        state_class: Type["BaseState"],
    ) -> None:
        super().__init__(name=name, types=types, instance=instance)
        self._state_class = state_class

    @property
    def state_class(self) -> Type["BaseState"]:
        return self._state_class

    @cached_property
    def id(self) -> UUID:
        """Generate deterministic UUID from state class and state value name."""
        return get_state_value_id(self._state_class, self.name)

    def resolve(self, state: "BaseState") -> _T:
        if hasattr(state, self._name):
            return cast(_T, getattr(state, self._name))

        if state.meta.parent:
            return self.resolve(state.meta.parent)

        raise NodeException(f"Missing required Workflow state: {self._name}", code=WorkflowErrorCode.INVALID_STATE)



from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.nodes.displayable.final_output_node import FinalOutputNode
from vellum.workflows.state.base import BaseState
from vellum.workflows.workflows.base import BaseWorkflow
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display


class Inputs(BaseInputs):
    input_value: str


class UnusedTerminalNode(FinalOutputNode):
    class Outputs(FinalOutputNode.Outputs):
        value = Inputs.input_value


class TerminalNodeInUnusedGraphsWorkflow(BaseWorkflow[Inputs, BaseState]):
    unused_graphs = {UnusedTerminalNode}

    class Outputs(BaseWorkflow.Outputs):
        result = UnusedTerminalNode.Outputs.value


def test_serialize_workflow__terminal_node_in_unused_graphs():
    """
    Tests that a workflow with a single terminal node in unused_graphs serializes correctly.
    """

    # GIVEN a Workflow with a single terminal node in unused_graphs
    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=TerminalNodeInUnusedGraphsWorkflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its output variables should contain the terminal node's output
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 1
    assert output_variables[0]["key"] == "result"
    assert output_variables[0]["type"] == "STRING"

    # AND its output values should reference the terminal node's output
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    output_values = workflow_raw_data["output_values"]
    assert len(output_values) == 1
    assert output_values[0]["output_variable_id"] == output_variables[0]["id"]
    assert output_values[0]["value"]["type"] == "NODE_OUTPUT"



"""Utilities for reading and managing workflow metadata.json files."""

import json
import os
from uuid import UUID
from typing import Any, Dict, Optional, Tuple, Type

from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.utils.files import virtual_open


def find_workflow_root_with_metadata(module_path: str) -> Optional[str]:
    """
    Find the workflow root module by searching for metadata.json up the module hierarchy.

    Args:
        module_path: The module path to search from (e.g., "workflows.my_workflow.nodes.my_node")

    Returns:
        The workflow root module path if found, None otherwise
    """
    parts = module_path.split(".")
    for i in range(len(parts), 0, -1):
        potential_root = ".".join(parts[:i])
        file_path = os.path.join(potential_root.replace(".", os.path.sep), "metadata.json")
        try:
            f = virtual_open(file_path)
            if f is not None:
                f.close()
                return potential_root
        except (FileNotFoundError, OSError):
            continue
    return None


def _load_workflow_metadata(module_path: str) -> Tuple[Optional[str], Optional[Dict]]:
    """
    Load the full metadata.json content for a given module.

    This function searches up the module hierarchy for metadata.json and loads its content.

    Args:
        module_path: The module path to search from (e.g., "workflows.my_workflow")

    Returns:
        A tuple of (workflow_root, metadata_dict). Returns (None, None) if not found.
    """
    try:
        workflow_root = find_workflow_root_with_metadata(module_path)
        if not workflow_root:
            return None, None
        file_path = os.path.join(workflow_root.replace(".", os.path.sep), "metadata.json")
        with virtual_open(file_path) as f:
            data = json.load(f)
            return workflow_root, data
    except Exception:
        return None, None


def _load_edges_mapping(module_path: str) -> Tuple[Optional[str], Dict[str, str]]:
    """
    Load edge path to ID mapping from metadata.json for a given module.

    This function searches up the module hierarchy for metadata.json and extracts
    the edges_to_id_mapping.

    Args:
        module_path: The module path to search from (e.g., "workflows.my_workflow")

    Returns:
        Dictionary mapping edge keys to their UUID strings
    """
    workflow_root, data = _load_workflow_metadata(module_path)
    if data is None:
        return None, {}
    edges_map = data.get("edges_to_id_mapping")
    if isinstance(edges_map, dict):
        return workflow_root, edges_map
    return workflow_root, {}


def load_edges_to_id_mapping(module_path: str) -> Dict[str, str]:
    _, mapping = _load_edges_mapping(module_path)
    return mapping


def _build_metadata_class_path(module: str, class_name: str, workflow_root: Optional[str]) -> str:
    """
    Format a class path so that it matches the metadata.json key structure.
    """
    if workflow_root:
        if module == workflow_root:
            return f".{class_name}"
        prefix = f"{workflow_root}."
        if module.startswith(prefix):
            relative = module[len(prefix) :]
            return f".{relative}.{class_name}" if relative else f".{class_name}"
    return f"{module}.{class_name}"


def get_trigger_edge_id(trigger_cls: Type, target_node: Type[BaseNode], module_path: str) -> Optional[str]:
    """
    Get the stable edge ID for a trigger edge from metadata.json.

    Args:
        trigger_cls: The trigger class
        target_node: The target node class
        module_path: The workflow module path

    Returns:
        The stable edge ID if found in metadata.json, None otherwise
    """
    workflow_root, edges_mapping = _load_edges_mapping(module_path)
    if not edges_mapping:
        return None
    trigger_path = _build_metadata_class_path(trigger_cls.__module__, trigger_cls.__name__, workflow_root)
    target_base = _build_metadata_class_path(target_node.__module__, target_node.__name__, workflow_root)
    target_path = f"{target_base}.Trigger"
    edge_key = f"{trigger_path}|{target_path}"
    return edges_mapping.get(edge_key)


def get_entrypoint_edge_id(target_node: Type[BaseNode], module_path: str) -> Optional[str]:
    """
    Get the stable edge ID for an entrypoint edge from metadata.json.

    Args:
        target_node: The target node class
        module_path: The workflow module path

    Returns:
        The stable edge ID if found in metadata.json, None otherwise
    """
    workflow_root, edges_mapping = _load_edges_mapping(module_path)
    if not edges_mapping:
        return None
    manual_path = "vellum.workflows.triggers.manual.Manual"
    target_base = _build_metadata_class_path(target_node.__module__, target_node.__name__, workflow_root)
    target_path = f"{target_base}.Trigger"
    edge_key = f"{manual_path}|{target_path}"
    return edges_mapping.get(edge_key)


def get_regular_edge_id(
    source_node_cls: Type[BaseNode], source_handle_id: UUID, target_node: Type[BaseNode], module_path: str
) -> Optional[str]:
    """
    Get the stable edge ID for a regular node-to-node edge from metadata.json.

    Args:
        source_node_cls: The source node class
        source_handle_id: The source port/handle ID
        target_node: The target node class
        module_path: The workflow module path

    Returns:
        The stable edge ID if found in metadata.json, None otherwise
    """
    workflow_root, edges_mapping = _load_edges_mapping(module_path)
    if not edges_mapping:
        return None
    source_base = _build_metadata_class_path(source_node_cls.__module__, source_node_cls.__name__, workflow_root)
    source_path = f"{source_base}.Ports.{str(source_handle_id)}"
    target_base = _build_metadata_class_path(target_node.__module__, target_node.__name__, workflow_root)
    target_path = f"{target_base}.Trigger"
    edge_key = f"{source_path}|{target_path}"
    return edges_mapping.get(edge_key)


def load_dataset_row_index_to_id_mapping(module_path: str) -> Dict[int, str]:
    """
    Load dataset row index to ID mapping from metadata.json for a given module.

    This function searches up the module hierarchy for metadata.json and extracts
    the dataset_row_index_to_id_mapping.

    Args:
        module_path: The module path to search from (e.g., "workflows.my_workflow")

    Returns:
        Dictionary mapping dataset row indices (as integers) to their ID strings
    """
    _, data = _load_workflow_metadata(module_path)
    if data is None:
        return {}
    mapping = data.get("dataset_row_index_to_id_mapping")
    if isinstance(mapping, dict):
        return {int(k): v for k, v in mapping.items()}
    return {}


def load_runner_config(module_path: str) -> Dict[str, Any]:
    """
    Load runner_config from metadata.json for a given module.

    This function searches up the module hierarchy for metadata.json and extracts
    the runner_config.

    Args:
        module_path: The module path to search from (e.g., "workflows.my_workflow")

    Returns:
        The runner_config dictionary if found in metadata.json, empty dict otherwise
    """
    _, data = _load_workflow_metadata(module_path)
    if data is None:
        return {}
    runner_config = data.get("runner_config")
    if isinstance(runner_config, dict):
        return runner_config
    return {}



from dataclasses import dataclass
from functools import cached_property
import json
import logging
from queue import Queue
import traceback
from uuid import UUID, uuid4
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type

from vellum import Vellum, __version__
from vellum.client.types import SeverityEnum
from vellum.workflows.context import ExecutionContext, get_execution_context, set_execution_context
from vellum.workflows.events.node import NodeExecutionLogBody, NodeExecutionLogEvent
from vellum.workflows.events.types import ExternalParentContext, NodeParentContext
from vellum.workflows.nodes.mocks import MockNodeExecution, MockNodeExecutionArg
from vellum.workflows.outputs.base import BaseOutputs
from vellum.workflows.references.constant import ConstantValueReference
from vellum.workflows.state.store import Store
from vellum.workflows.utils.uuids import generate_workflow_deployment_prefix
from vellum.workflows.utils.zip import extract_zip_files
from vellum.workflows.vellum_client import create_vellum_client

if TYPE_CHECKING:
    from vellum.workflows.events.workflow import WorkflowEvent
    from vellum.workflows.state.base import BaseState
    from vellum.workflows.workflows.base import BaseWorkflow

logger = logging.getLogger(__name__)


@dataclass
class WorkflowDeploymentMetadata:
    """Metadata about a workflow deployment needed for parent context construction."""

    deployment_id: UUID
    deployment_name: str
    deployment_history_item_id: UUID
    release_tag_id: UUID
    release_tag_name: str
    workflow_version_id: UUID


class WorkflowContext:
    def __init__(
        self,
        *,
        vellum_client: Optional[Vellum] = None,
        execution_context: Optional[ExecutionContext] = None,
        generated_files: Optional[dict[str, str]] = None,
        namespace: Optional[str] = None,
        store_class: Optional[Type[Store]] = None,
        event_max_size: Optional[int] = None,
    ):
        self._vellum_client = vellum_client
        self._event_queue: Optional[Queue["WorkflowEvent"]] = None
        self._node_output_mocks_map: Dict[Type[BaseOutputs], List[MockNodeExecution]] = {}
        self._execution_context = get_execution_context()
        self._namespace = namespace
        self._store_class = store_class if store_class is not None else Store
        self._event_max_size = event_max_size

        if execution_context is not None:
            self._execution_context.trace_id = execution_context.trace_id
            if execution_context.parent_context is not None:
                self._execution_context.parent_context = execution_context.parent_context

        if self._execution_context.parent_context is None:
            self._execution_context.parent_context = ExternalParentContext(span_id=uuid4())
            # Only generate a new trace_id if one wasn't explicitly provided (i.e., if it's the default zero UUID)
            if self._execution_context.trace_id == UUID("00000000-0000-0000-0000-000000000000"):
                self._execution_context.trace_id = uuid4()
            # Propagate the updated context back to the global execution context
            set_execution_context(self._execution_context)

        self._generated_files = generated_files

    @cached_property
    def vellum_client(self) -> Vellum:
        if self._vellum_client:
            return self._vellum_client

        return create_vellum_client()

    @cached_property
    def execution_context(self) -> ExecutionContext:
        return self._execution_context

    @cached_property
    def generated_files(self) -> Optional[dict[str, str]]:
        return self._generated_files

    @cached_property
    def namespace(self) -> Optional[str]:
        return self._namespace

    @cached_property
    def node_output_mocks_map(self) -> Dict[Type[BaseOutputs], List[MockNodeExecution]]:
        return self._node_output_mocks_map

    @property
    def store_class(self) -> Type[Store]:
        return self._store_class

    @property
    def event_max_size(self) -> Optional[int]:
        return self._event_max_size

    @property
    def monitoring_url(self) -> Optional[str]:
        """
        Get the base monitoring URL for this workflow context.

        Returns the base URL to view executions in Vellum UI.
        """
        # Get UI URL from the Vellum client's API URL
        api_url = self.vellum_client._client_wrapper.get_environment().default
        return self._get_ui_url_from_api_url(api_url)

    def get_monitoring_url(self, span_id: str) -> Optional[str]:
        """
        Generate the monitoring URL for this workflow execution.

        Args:
        span_id: The span ID from the workflow execution result.

        Returns:
        The URL to view execution details in Vellum UI, or None if monitoring is disabled.
        """
        base_url = self.monitoring_url
        if base_url is None:
            return None

        return f"{base_url}/workflows/executions/{span_id}"

    def _get_ui_url_from_api_url(self, api_url: str) -> str:
        """
        Convert an API URL to the corresponding UI URL.

        Args:
        api_url: The API base URL (e.g., https://api.vellum.ai or http://localhost:8000)

        Returns:
        The corresponding UI URL (e.g., https://app.vellum.ai or http://localhost:3000)
        """
        if "localhost" in api_url:
            # For local development: localhost:8000 (API) -> localhost:3000 (UI)
            return api_url.replace(":8000", ":3000")
        elif "api.vellum.ai" in api_url:
            # For production: api.vellum.ai -> app.vellum.ai
            return api_url.replace("api.vellum.ai", "app.vellum.ai")
        else:
            # For custom domains, assume the same pattern: api.* -> app.*
            return api_url.replace("api.", "app.", 1)

    def emit_log_event(
        self,
        severity: SeverityEnum,
        message: str,
        attributes: Optional[Dict[str, Any]] = None,
        exc_info: Optional[bool] = None,
    ) -> None:
        """Emit a log event for a particular node.

        This is in active development and may have breaking changes.
        """
        from vellum.workflows.nodes.bases import BaseNode

        if self._event_queue is None:
            return

        execution_context = get_execution_context()
        parent_context = execution_context.parent_context
        while parent_context:
            if isinstance(parent_context, NodeParentContext):
                break
            parent_context = parent_context.parent

        if not isinstance(parent_context, NodeParentContext):
            return

        try:
            node_class = parent_context.node_definition.decode()
        except Exception:
            logger.exception("Failed to decode node definition.")
            return

        if not isinstance(node_class, type) or not issubclass(node_class, BaseNode):
            logger.warning("Node definition is not a subclass of BaseNode.")
            return

        if exc_info:
            attributes = {
                **(attributes or {}),
                "exc_info": traceback.format_exc(),
            }

        self._event_queue.put(
            NodeExecutionLogEvent(
                trace_id=execution_context.trace_id,
                span_id=parent_context.span_id,
                body=NodeExecutionLogBody(
                    node_definition=node_class,
                    severity=severity,
                    message=message,
                    attributes=attributes,
                ),
            )
        )

    def _emit_subworkflow_event(self, event: "WorkflowEvent") -> None:
        if self._event_queue:
            self._event_queue.put(event)

    def _register_event_queue(self, event_queue: Queue["WorkflowEvent"]) -> None:
        self._event_queue = event_queue

    def _register_event_max_size(self, event_max_size: Optional[int]) -> None:
        self._event_max_size = event_max_size

    def _register_node_output_mocks(self, node_output_mocks: MockNodeExecutionArg) -> None:
        for mock in node_output_mocks:
            if isinstance(mock, MockNodeExecution):
                key = mock.then_outputs.__class__
                value = mock
            else:
                key = mock.__class__
                value = MockNodeExecution(
                    when_condition=ConstantValueReference(True),
                    then_outputs=mock,
                )

            if key not in self._node_output_mocks_map:
                self._node_output_mocks_map[key] = [value]
            else:
                self._node_output_mocks_map[key].append(value)

    def _get_all_node_output_mocks(self) -> List[MockNodeExecution]:
        return [mock for mocks in self._node_output_mocks_map.values() for mock in mocks]

    def resolve_workflow_deployment(
        self, deployment_name: str, release_tag: str, state: "BaseState"
    ) -> Optional[Tuple[Type["BaseWorkflow"], Optional[WorkflowDeploymentMetadata]]]:
        """
        Resolve a workflow deployment by name and release tag.

        Args:
            deployment_name: The name of the workflow deployment
            release_tag: The release tag to resolve
            state: The base state to pass to the workflow

        Returns:
            Tuple of (BaseWorkflow class, deployment metadata) if found
        """
        if not self._generated_files or not self._namespace:
            return None

        expected_prefix = generate_workflow_deployment_prefix(deployment_name, release_tag)
        WorkflowClass: Optional[Type["BaseWorkflow"]] = None

        try:
            from vellum.workflows.workflows.base import BaseWorkflow

            WorkflowClass = BaseWorkflow.load_from_module(f"{self.namespace}.{expected_prefix}")
            WorkflowClass.is_dynamic = True
        except Exception:
            pass

        if WorkflowClass is None:
            try:
                major_version = __version__.split(".")[0]
                version_range = f">={major_version}.0.0,<={__version__}"

                response = self.vellum_client.workflows.pull(
                    deployment_name,
                    release_tag=release_tag,
                    version=version_range,
                    request_options={"additional_headers": {"X-Vellum-Always-Success": "true"}},
                )

                if isinstance(response, dict) and response.get("success") is False:
                    return None

                zip_bytes = b"".join(response)
                pulled_files = extract_zip_files(zip_bytes)

                for file_name, content in pulled_files.items():
                    prefixed_file_name = f"{expected_prefix}/{file_name}"
                    self._generated_files[prefixed_file_name] = content

                from vellum.workflows.workflows.base import BaseWorkflow

                WorkflowClass = BaseWorkflow.load_from_module(f"{self.namespace}.{expected_prefix}")
                WorkflowClass.is_dynamic = True

            except Exception:
                pass

        # If we successfully loaded the workflow class, fetch metadata from generated_files (important-comment)
        # This ensures we pick up metadata.json if a pull just populated it
        if WorkflowClass is not None:
            deployment_metadata = self._fetch_deployment_metadata(deployment_name, release_tag)
            return (WorkflowClass, deployment_metadata)

        return None

    def _fetch_deployment_metadata(
        self, deployment_name: str, release_tag: str
    ) -> Optional[WorkflowDeploymentMetadata]:
        """
        Fetch deployment metadata from metadata.json in generated_files.

        Args:
            deployment_name: The name of the workflow deployment
            release_tag: The release tag name

        Returns:
            WorkflowDeploymentMetadata if successful, None otherwise
        """
        if not self._generated_files:
            return None

        expected_prefix = generate_workflow_deployment_prefix(deployment_name, release_tag)
        metadata_path = f"{expected_prefix}/metadata.json"

        metadata_content = self._generated_files.get(metadata_path)
        if metadata_content:
            try:
                metadata_json = json.loads(metadata_content)

                deployment_id = metadata_json.get("deployment_id")
                deployment_name_from_metadata = metadata_json.get("deployment_name")
                deployment_history_item_id = metadata_json.get("deployment_history_item_id")
                release_tag_id = metadata_json.get("release_tag_id")
                release_tag_name_from_metadata = metadata_json.get("release_tag_name")
                workflow_version_id = metadata_json.get("workflow_version_id")

                if all(
                    [
                        deployment_id,
                        deployment_name_from_metadata,
                        deployment_history_item_id,
                        release_tag_id,
                        release_tag_name_from_metadata,
                        workflow_version_id,
                    ]
                ):
                    return WorkflowDeploymentMetadata(
                        deployment_id=UUID(deployment_id),
                        deployment_name=deployment_name_from_metadata,
                        deployment_history_item_id=UUID(deployment_history_item_id),
                        release_tag_id=UUID(release_tag_id),
                        release_tag_name=release_tag_name_from_metadata,
                        workflow_version_id=UUID(workflow_version_id),
                    )
            except (json.JSONDecodeError, ValueError, KeyError):
                # If we fail to parse metadata, return None - the workflow can still run
                # but won't have the full parent context hierarchy
                pass

        return None

    @classmethod
    def create_from(cls, context: "WorkflowContext") -> "WorkflowContext":
        return cls(
            vellum_client=context.vellum_client,
            generated_files=context.generated_files,
            namespace=context.namespace,
            store_class=context.store_class,
            event_max_size=context.event_max_size,
        )



from vellum.workflows import BaseWorkflow
from vellum.workflows.nodes.displayable.note_node.node import NoteNode
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display


def test_serialize_node__note_node():
    """
    Tests that a single note node is properly serialized in a workflow.
    """

    # GIVEN a note node with text and style defined on the node class
    class MyNoteNode(NoteNode):
        text = "This makes sense"
        style = {
            "fontSize": 24,
        }

    # AND a workflow with the note node
    class Workflow(BaseWorkflow):
        graph = MyNoteNode

    # WHEN the workflow is serialized
    workflow_display = get_workflow_display(workflow_class=Workflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN the node should properly serialize the inputs
    my_note_node = next(node for node in serialized_workflow["workflow_raw_data"]["nodes"] if node["type"] == "NOTE")

    assert my_note_node["inputs"] == []
    assert my_note_node["data"]["text"] == "This makes sense"
    assert my_note_node["data"]["style"] == {"fontSize": 24}


def test_serialize_workflow__graph_with_only_note_nodes():
    """
    Tests that a workflow with only note nodes properly serializes all note nodes in the final payload.
    """

    # GIVEN multiple note nodes with text and style defined on the node classes
    class FirstNoteNode(NoteNode):
        text = "First note"
        style = {"fontSize": 16}

    class SecondNoteNode(NoteNode):
        text = "Second note"
        style = {"fontSize": 20}

    # AND a workflow with only note nodes in unused_graphs
    class Workflow(BaseWorkflow):
        unused_graphs = {FirstNoteNode, SecondNoteNode}  # type: ignore[assignment]

    # WHEN the workflow is serialized
    workflow_display = get_workflow_display(workflow_class=Workflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN the serialized workflow should contain both note nodes
    note_nodes = [node for node in serialized_workflow["workflow_raw_data"]["nodes"] if node["type"] == "NOTE"]
    assert len(note_nodes) == 2

    # AND the note nodes should have the correct data
    note_texts = {node["data"]["text"] for node in note_nodes}
    assert note_texts == {"First note", "Second note"}

    # AND each note node should have the correct style
    first_note = next(node for node in note_nodes if node["data"]["text"] == "First note")
    second_note = next(node for node in note_nodes if node["data"]["text"] == "Second note")

    assert first_note["data"]["style"] == {"fontSize": 16}
    assert second_note["data"]["style"] == {"fontSize": 20}



import pytest
from copy import deepcopy
import json
from queue import Queue
import threading
from typing import Any, Dict, List, Optional, cast

from pydantic import Field

from vellum import ChatMessage
from vellum.utils.json_encoder import VellumJsonEncoder
from vellum.workflows.constants import undefined
from vellum.workflows.nodes.bases import BaseNode
from vellum.workflows.outputs.base import BaseOutputs
from vellum.workflows.state.base import BaseState
from vellum.workflows.state.delta import SetStateDelta, StateDelta
from vellum.workflows.types.code_execution_node_wrappers import DictWrapper


@pytest.fixture()
def mock_deepcopy(mocker):
    return mocker.patch("vellum.workflows.state.base.deepcopy")


@pytest.fixture
def mock_logger(mocker):
    return mocker.patch("vellum.workflows.state.base.logger")


class MockState(BaseState):
    foo: str
    nested_dict: Dict[str, int] = {}

    __snapshot_count__: int = 0

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.__snapshot_callback__ = lambda _, __: self.__mock_snapshot__()

    def __mock_snapshot__(self) -> None:
        self.__snapshot_count__ += 1

    def __deepcopy__(self, memo: dict) -> "MockState":
        new_state = cast(MockState, super().__deepcopy__(memo))
        new_state.__snapshot_count__ = 0
        return new_state


class MockNode(BaseNode):
    class ExternalInputs(BaseNode.ExternalInputs):
        message: str

    class Outputs(BaseOutputs):
        baz: str


MOCK_NODE_OUTPUT_ID = "a864331d-9b1c-43db-90e7-6b1998d8179d"


def test_state_snapshot__node_attribute_edit():
    # GIVEN an initial state instance
    state = MockState(foo="bar")
    assert state.__snapshot_count__ == 0

    # WHEN we edit an attribute
    state.foo = "baz"

    # THEN the snapshot is emitted
    assert state.__snapshot_count__ == 1


def test_state_snapshot__node_output_edit():
    # GIVEN an initial state instance
    state = MockState(foo="bar")
    assert state.__snapshot_count__ == 0

    # WHEN we add a Node Output to state
    for output in MockNode.Outputs:
        state.meta.node_outputs[output] = "hello"

    # THEN the snapshot is emitted
    assert state.__snapshot_count__ == 1


def test_state_snapshot__nested_dictionary_edit():
    # GIVEN an initial state instance
    state = MockState(foo="bar")
    assert state.__snapshot_count__ == 0

    # WHEN we edit a nested dictionary
    state.nested_dict["hello"] = 1

    # THEN the snapshot is emitted
    assert state.__snapshot_count__ == 1


def test_state_snapshot__external_input_edit():
    # GIVEN an initial state instance
    state = MockState(foo="bar")
    assert state.__snapshot_count__ == 0

    # WHEN we add an external input to state
    state.meta.external_inputs[MockNode.ExternalInputs.message] = "hello"

    # THEN the snapshot is emitted
    assert state.__snapshot_count__ == 1


def test_state_deepcopy():
    # GIVEN an initial state instance
    state = MockState(foo="bar")

    # AND we add a Node Output to state
    state.meta.node_outputs[MockNode.Outputs.baz] = "hello"

    # WHEN we deepcopy the state
    deepcopied_state = deepcopy(state)

    # THEN node outputs are deepcopied
    assert deepcopied_state.meta.node_outputs == state.meta.node_outputs


def test_state_deepcopy__with_node_output_updates():
    # GIVEN an initial state instance
    state = MockState(foo="bar")

    # AND we add a Node Output to state
    state.meta.node_outputs[MockNode.Outputs.baz] = "hello"

    # AND we deepcopy the state
    deepcopied_state = deepcopy(state)

    # AND we update the original state
    state.meta.node_outputs[MockNode.Outputs.baz] = "world"

    # THEN the copied state is not updated
    assert deepcopied_state.meta.node_outputs[MockNode.Outputs.baz] == "hello"

    # AND the original state has had the correct number of snapshots
    assert state.__snapshot_count__ == 2

    # AND the copied state has had the correct number of snapshots
    assert deepcopied_state.__snapshot_count__ == 0


def test_state_json_serialization__with_node_output_updates():
    # GIVEN an initial state instance
    state = MockState(foo="bar")

    # AND we add a Node Output to state
    state.meta.node_outputs[MockNode.Outputs.baz] = "hello"

    # WHEN we serialize the state
    json_state = json.loads(json.dumps(state, cls=VellumJsonEncoder))

    # THEN the state is serialized correctly
    assert json_state["meta"]["node_outputs"] == {MOCK_NODE_OUTPUT_ID: "hello"}


def test_state_deepcopy__with_external_input_updates():
    # GIVEN an initial state instance
    state = MockState(foo="bar")

    # AND we add an external input to state
    state.meta.external_inputs[MockNode.ExternalInputs.message] = "hello"

    # AND we deepcopy the state
    deepcopied_state = deepcopy(state)

    # AND we update the original state
    state.meta.external_inputs[MockNode.ExternalInputs.message] = "world"

    # THEN the copied state is not updated
    assert deepcopied_state.meta.external_inputs[MockNode.ExternalInputs.message] == "hello"

    # AND the original state has had the correct number of snapshots
    assert state.__snapshot_count__ == 2

    # AND the copied state has had the correct number of snapshots
    assert deepcopied_state.__snapshot_count__ == 0


def test_state_json_serialization__with_queue():
    # GIVEN an initial state instance
    state = MockState(foo="bar")

    # AND we add a Node Output queue to state
    queue: Queue[str] = Queue()
    queue.put("test1")
    queue.put("test2")

    state.meta.node_outputs[MockNode.Outputs.baz] = queue

    # WHEN we serialize the state
    json_state = json.loads(json.dumps(state, cls=VellumJsonEncoder))

    # THEN the state is serialized correctly with the queue turned into a list
    assert json_state["meta"]["node_outputs"] == {MOCK_NODE_OUTPUT_ID: ["test1", "test2"]}


def test_state_snapshot__deepcopy_fails__logs_error(mock_deepcopy, mock_logger):
    # GIVEN an initial state instance
    class MockState(BaseState):
        foo: str

    state = MockState(foo="bar")

    # AND we have a snapshot callback that we are tracking
    snapshot_count = 0

    def __snapshot_callback__(state: "BaseState", deltas: List[StateDelta]) -> None:
        nonlocal snapshot_count
        snapshot_count += 1

    state.__snapshot_callback__ = __snapshot_callback__

    # AND deepcopy fails the first time but succeeds the second time
    deepcopy_side_effect_count = 0

    def side_effect(*args, **kwargs):
        nonlocal deepcopy_side_effect_count
        deepcopy_side_effect_count += 1
        if deepcopy_side_effect_count == 1:
            raise Exception("Failed to deepcopy")
        return deepcopy(args[0])

    mock_deepcopy.side_effect = side_effect

    # WHEN we snapshot the state twice
    state.__snapshot__(SetStateDelta(name="foo", delta="bar"))
    state.__snapshot__(SetStateDelta(name="foo", delta="baz"))

    # THEN we were able to invoke the callback once
    assert snapshot_count == 1

    # AND alert sentry once
    assert mock_logger.exception.call_count == 1


def test_state_deepcopy_handles_undefined_values():
    # GIVEN a state with undefined values in node outputs
    state = MockState(foo="bar")
    state.meta.node_outputs[MockNode.Outputs.baz] = DictWrapper({"foo": undefined})

    # WHEN we deepcopy the state
    deepcopied_state = deepcopy(state)

    # THEN the undefined values are preserved
    assert deepcopied_state.meta.node_outputs[MockNode.Outputs.baz] == {"foo": undefined}


def test_base_state_initializes_field_with_default_factory():
    """Test that BaseState properly initializes fields with Field(default_factory=...)."""

    # GIVEN a state class with fields using Field(default_factory=...)
    class TestState(BaseState):
        chat_history: List[str] = Field(default_factory=list)
        items: Dict[str, int] = Field(default_factory=dict)
        counter: int = Field(default_factory=lambda: 0)

    # WHEN we create a state instance without providing values
    state = TestState()

    # THEN the fields should be initialized with the factory results, not FieldInfo objects
    assert isinstance(state.chat_history, list)
    assert state.chat_history == []
    assert isinstance(state.items, dict)
    assert state.items == {}
    assert isinstance(state.counter, int)
    assert state.counter == 0

    # AND we should be able to modify them
    state.chat_history.append("message1")
    state.items["key1"] = 1
    state.counter += 1

    assert state.chat_history == ["message1"]
    assert state.items == {"key1": 1}
    assert state.counter == 1


def test_base_state_field_with_default_factory_creates_separate_instances():
    """Test that Field(default_factory=...) creates separate instances for each state."""

    # GIVEN a state class with Field(default_factory=list)
    class TestState(BaseState):
        items: List[str] = Field(default_factory=list)

    # WHEN we create two state instances
    state1 = TestState()
    state2 = TestState()

    # THEN they should have separate list instances
    assert state1.items is not state2.items

    # AND modifying one should not affect the other
    state1.items.append("item1")
    assert state1.items == ["item1"]
    assert state2.items == []


class BlockingValue:
    """A value that blocks during deepcopy until signaled to proceed."""

    def __init__(self, entered_event: threading.Event, proceed_event: threading.Event):
        self.entered_event = entered_event
        self.proceed_event = proceed_event

    def __deepcopy__(self, memo: Any) -> "BlockingValue":
        self.entered_event.set()
        self.proceed_event.wait(timeout=5.0)
        return BlockingValue(self.entered_event, self.proceed_event)


def test_state_snapshot__concurrent_mutation_during_deepcopy():
    """Test that concurrent mutations during deepcopy don't cause RuntimeError."""

    # GIVEN a state with a dict containing a blocking value
    class TestState(BaseState):
        data: Dict[str, Any] = Field(default_factory=dict)

    state = TestState()

    entered_event = threading.Event()
    proceed_event = threading.Event()
    state.data["blocking"] = BlockingValue(entered_event, proceed_event)
    state.data["other"] = "value"

    snapshot_exception: List[Exception] = []
    mutation_completed = threading.Event()

    def snapshot_thread_fn() -> None:
        try:
            with state.__lock__:
                deepcopy(state)
        except Exception as e:
            snapshot_exception.append(e)

    def mutation_thread_fn() -> None:
        state.data["new_key"] = "new_value"
        mutation_completed.set()

    # WHEN we start a snapshot (deepcopy) in one thread
    snapshot_thread = threading.Thread(target=snapshot_thread_fn)
    snapshot_thread.start()

    # AND wait for the deepcopy to be in progress (blocked on our blocking value)
    entered_event.wait(timeout=5.0)

    # AND try to mutate the dict from another thread
    mutation_thread = threading.Thread(target=mutation_thread_fn)
    mutation_thread.start()

    # THEN the mutation should block waiting for the lock (not complete immediately)
    mutation_completed.wait(timeout=0.2)
    mutation_blocked = not mutation_completed.is_set()

    # AND when we allow the deepcopy to proceed
    proceed_event.set()
    snapshot_thread.join(timeout=5.0)
    mutation_thread.join(timeout=5.0)

    # THEN the mutation should have been blocked by the lock
    assert mutation_blocked, "Mutation should block while deepcopy holds the lock"

    # AND no exception should have been raised during snapshot
    assert len(snapshot_exception) == 0, f"Snapshot raised exception: {snapshot_exception}"


def test_state_deepcopy__cloned_state_uses_own_snapshot_callback():
    """Test that deepcopied state's snapshottable containers use the clone's callback."""

    # GIVEN a state with a snapshottable dict attribute
    original_snapshot_count = 0
    clone_snapshot_count = 0

    class TestState(BaseState):
        data: Dict[str, int] = Field(default_factory=dict)

    state = TestState()
    state.data["key1"] = 1

    def original_callback(state_copy: BaseState, deltas: List[StateDelta]) -> None:
        nonlocal original_snapshot_count
        original_snapshot_count += 1

    state.__snapshot_callback__ = original_callback

    # WHEN we deepcopy the state
    cloned_state = deepcopy(state)

    def clone_callback(state_copy: BaseState, deltas: List[StateDelta]) -> None:
        nonlocal clone_snapshot_count
        clone_snapshot_count += 1

    cloned_state.__snapshot_callback__ = clone_callback

    # AND reset counters
    original_snapshot_count = 0
    clone_snapshot_count = 0

    # AND mutate the cloned state's snapshottable dict
    cloned_state.data["key2"] = 2

    # THEN only the clone's callback should be invoked
    assert clone_snapshot_count == 1, "Clone's callback should be invoked"
    assert original_snapshot_count == 0, "Original's callback should not be invoked"


def test_state_snapshot__top_level_attribute_assignment_blocks_during_deepcopy():
    """Test that top-level attribute assignments block while deepcopy holds the lock."""

    # GIVEN a state with a blocking value in a dict attribute
    class TestState(BaseState):
        data: Dict[str, Any] = Field(default_factory=dict)
        counter: int = 0

    state = TestState()

    entered_event = threading.Event()
    proceed_event = threading.Event()
    state.data["blocking"] = BlockingValue(entered_event, proceed_event)

    mutation_completed = threading.Event()

    def snapshot_thread_fn() -> None:
        with state.__lock__:
            deepcopy(state)

    def mutation_thread_fn() -> None:
        state.__is_quiet__ = True
        state.counter = 42
        mutation_completed.set()

    # WHEN we start a snapshot (deepcopy) in one thread
    snapshot_thread = threading.Thread(target=snapshot_thread_fn)
    snapshot_thread.start()

    # AND wait for the deepcopy to be in progress (blocked on our blocking value)
    entered_event.wait(timeout=5.0)

    # AND try to assign a top-level attribute from another thread
    mutation_thread = threading.Thread(target=mutation_thread_fn)
    mutation_thread.start()

    # THEN the mutation should block waiting for the lock (not complete immediately)
    mutation_completed.wait(timeout=0.2)
    mutation_blocked = not mutation_completed.is_set()

    # AND when we allow the deepcopy to proceed
    proceed_event.set()
    snapshot_thread.join(timeout=5.0)
    mutation_thread.join(timeout=5.0)

    # THEN the mutation should have been blocked by the lock
    assert mutation_blocked, "Top-level attribute assignment should block while deepcopy holds the lock"


def test_base_state_chat_history_with_default_factory_initializes_to_list():
    """
    Tests that a chat_history state variable with Optional[list[ChatMessage]] = Field(default_factory=list)
    initializes to an empty list instead of None.
    """

    # GIVEN a state class with chat_history using Field(default_factory=list)
    class TestState(BaseState):
        chat_history: Optional[List[ChatMessage]] = Field(default_factory=list)  # type: ignore[arg-type]

    # WHEN we create a state instance without providing a value
    state = TestState()

    # THEN the chat_history should be an empty list, not None
    assert state.chat_history is not None
    assert isinstance(state.chat_history, list)
    assert state.chat_history == []

    # AND we should be able to append ChatMessage objects to it
    chat_history = state.chat_history
    chat_history.append(ChatMessage(role="USER", text="Hello"))
    assert len(chat_history) == 1
    assert chat_history[0].role == "USER"
    assert chat_history[0].text == "Hello"


def test_base_state_chat_history_with_default_factory_creates_separate_instances():
    """
    Tests that Field(default_factory=list) creates separate list instances for each state,
    avoiding the mutable default argument issue.
    """

    # GIVEN a state class with chat_history using Field(default_factory=list)
    class TestState(BaseState):
        chat_history: Optional[List[ChatMessage]] = Field(default_factory=list)  # type: ignore[arg-type]

    # WHEN we create two state instances
    state1 = TestState()
    state2 = TestState()

    # THEN they should have separate list instances
    assert state1.chat_history is not state2.chat_history

    # AND modifying one should not affect the other
    chat_history1 = state1.chat_history
    chat_history2 = state2.chat_history
    assert chat_history1 is not None
    assert chat_history2 is not None
    chat_history1.append(ChatMessage(role="USER", text="Message 1"))
    chat_history2.append(ChatMessage(role="ASSISTANT", text="Message 2"))

    assert len(chat_history1) == 1
    assert len(chat_history2) == 1
    assert chat_history1[0].text == "Message 1"
    assert chat_history2[0].text == "Message 2"



import logging
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Sequence, Type, Union

from pydantic import ConfigDict, SerializationInfo, ValidationError, field_serializer, model_serializer

from vellum.client.core.pydantic_utilities import UniversalBaseModel
from vellum.client.types.array_vellum_value import ArrayVellumValue
from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.events.types import default_serializer
from vellum.workflows.exceptions import NodeException, WorkflowInitializationException
from vellum.workflows.nodes.utils import get_unadorned_node
from vellum.workflows.outputs.base import BaseOutputs
from vellum.workflows.references.constant import ConstantValueReference

if TYPE_CHECKING:
    from vellum.workflows import BaseWorkflow

logger = logging.getLogger(__name__)


class MockNodeExecution(UniversalBaseModel):
    when_condition: BaseDescriptor
    then_outputs: BaseOutputs
    disabled: Optional[bool] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @model_serializer(mode="wrap")
    def serialize_full_model(self, handler: Callable[[Any], Any], info: SerializationInfo) -> Dict[str, Any]:
        """Serialize the model and add node_id field computed from then_outputs."""
        serialized = handler(self)
        node_class = self.then_outputs.__class__.__parent_class__
        unadorned_node = get_unadorned_node(node_class)  # type: ignore[arg-type]
        serialized["node_id"] = str(unadorned_node.__id__)
        serialized["type"] = "NODE_EXECUTION"
        if self.disabled is None:
            del serialized["disabled"]
        return serialized

    @field_serializer("then_outputs")
    def serialize_then_outputs(self, then_outputs: BaseOutputs, _info: SerializationInfo) -> Dict[str, Any]:
        return default_serializer(then_outputs)

    @field_serializer("when_condition")
    def serialize_when_condition(self, when_condition: BaseDescriptor, info: SerializationInfo) -> Dict[str, Any]:
        # This allows `ee` to pass in workflow display context through the pydantic context
        if isinstance(info.context, dict) and "serializer" in info.context and callable(info.context["serializer"]):
            return info.context["serializer"](when_condition)

        return default_serializer(when_condition)

    @staticmethod
    def validate_all(
        raw_mock_workflow_node_configs: Optional[List[Any]],
        workflow: Type["BaseWorkflow"],
        descriptor_validator: Optional[Callable[[dict, Type["BaseWorkflow"]], BaseDescriptor]] = None,
    ) -> Optional[List["MockNodeExecution"]]:
        if not raw_mock_workflow_node_configs:
            return None

        ArrayVellumValue.model_rebuild()
        mock_node_executions: list[MockNodeExecution] = []
        for raw_mock_workflow_node_config in raw_mock_workflow_node_configs:
            if not isinstance(raw_mock_workflow_node_config, dict):
                raise WorkflowInitializationException(
                    message=f"Invalid mock node execution type: {type(raw_mock_workflow_node_config)}",
                    code=WorkflowErrorCode.INVALID_INPUTS,
                    workflow_definition=workflow,
                )

            try:
                mock_node_executions.append(
                    MockNodeExecution.model_validate(
                        raw_mock_workflow_node_config,
                        context={
                            "workflow": workflow,
                            "node_id": raw_mock_workflow_node_config.get("node_id"),
                            "descriptor_validator": lambda value: (
                                descriptor_validator(value, workflow)
                                if descriptor_validator
                                else ConstantValueReference(False)
                            ),
                        },
                    )
                )
            except WorkflowInitializationException as e:
                # If the node is not found in the workflow, skip it with a warning
                node_id = raw_mock_workflow_node_config.get("node_id")
                raw_data = e.raw_data or {}
                if raw_data.get("node_ref") == node_id:
                    logger.warning(
                        "Skipping mock for node %s: node not found in workflow %s",
                        node_id,
                        workflow.__name__,
                    )
                    continue
                raise
            except (ValidationError, NodeException) as e:
                raise WorkflowInitializationException(
                    message="Failed to validate mock node executions",
                    code=WorkflowErrorCode.INVALID_INPUTS,
                    workflow_definition=workflow,
                ) from e

        return mock_node_executions


MockNodeExecutionArg = Sequence[Union[BaseOutputs, MockNodeExecution]]



import pytest
from unittest.mock import patch
from uuid import UUID, uuid4
from typing import List

from vellum.client.types.string_vellum_value import StringVellumValue
from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.inputs import BaseInputs
from vellum.workflows.nodes.bases import BaseNode
from vellum.workflows.outputs import BaseOutputs
from vellum.workflows.references import LazyReference
from vellum.workflows.references.state_value import StateValueReference
from vellum.workflows.references.trigger import TriggerAttributeReference
from vellum.workflows.state import BaseState
from vellum.workflows.triggers.base import BaseTrigger
from vellum_ee.workflows.display.base import StateValueDisplay, WorkflowInputsDisplay, WorkflowMetaDisplay
from vellum_ee.workflows.display.editor.types import NodeDisplayData
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay
from vellum_ee.workflows.display.nodes.types import NodeOutputDisplay
from vellum_ee.workflows.display.nodes.vellum.utils import create_node_input, create_node_input_value_pointer_rules
from vellum_ee.workflows.display.types import WorkflowDisplayContext
from vellum_ee.workflows.display.utils.vellum import (
    ConstantValuePointer,
    InputVariableData,
    InputVariablePointer,
    NodeInputValuePointerRule,
    NodeOutputData,
    NodeOutputPointer,
    TriggerAttributeData,
    TriggerAttributePointer,
    WorkflowStateData,
    WorkflowStatePointer,
    create_node_input_value_pointer_rule,
)
from vellum_ee.workflows.display.workflows.base_workflow_display import BaseWorkflowDisplay


class Inputs(BaseInputs):
    example_workflow_input: str


class MyNodeA(BaseNode):
    example_node_input = Inputs.example_workflow_input

    class Outputs(BaseOutputs):
        output: str


class MyNodeADisplay(BaseNodeDisplay[MyNodeA]):
    pass


class MyNodeB(BaseNode):
    example = MyNodeA.Outputs.output
    fallback_example = MyNodeA.Outputs.output.coalesce(Inputs.example_workflow_input).coalesce("fallback")
    constant_coalesce = Inputs.example_workflow_input.coalesce("default_value")
    lazy_coalesce: BaseDescriptor[str] = LazyReference(
        lambda: MyNodeA.Outputs.output.coalesce(Inputs.example_workflow_input)
    )


@pytest.mark.parametrize(
    ["descriptor", "expected_rules"],
    [
        (
            MyNodeB.example,
            [
                NodeOutputPointer(
                    data=NodeOutputData(
                        node_id="f2f8f47a-25de-474c-95d7-df4f9b4cb5ac",
                        output_id="4b16a629-11a1-4b3f-a965-a57b872d13b8",
                    )
                )
            ],
        ),
        (
            MyNodeB.fallback_example,
            [
                NodeOutputPointer(
                    data=NodeOutputData(
                        node_id="f2f8f47a-25de-474c-95d7-df4f9b4cb5ac",
                        output_id="4b16a629-11a1-4b3f-a965-a57b872d13b8",
                    ),
                ),
                InputVariablePointer(
                    data=InputVariableData(input_variable_id="a154c29d-fac0-4cd0-ba88-bc52034f5470"),
                ),
                ConstantValuePointer(data=StringVellumValue(value="fallback")),
            ],
        ),
        (
            MyNodeB.constant_coalesce,
            [
                InputVariablePointer(
                    data=InputVariableData(input_variable_id="a154c29d-fac0-4cd0-ba88-bc52034f5470"),
                ),
                ConstantValuePointer(data=StringVellumValue(value="default_value")),
            ],
        ),
        (
            MyNodeB.lazy_coalesce,
            [
                NodeOutputPointer(
                    data=NodeOutputData(
                        node_id="f2f8f47a-25de-474c-95d7-df4f9b4cb5ac",
                        output_id="4b16a629-11a1-4b3f-a965-a57b872d13b8",
                    ),
                ),
                InputVariablePointer(
                    data=InputVariableData(input_variable_id="a154c29d-fac0-4cd0-ba88-bc52034f5470"),
                ),
            ],
        ),
    ],
)
def test_create_node_input_value_pointer_rules(
    descriptor: BaseDescriptor, expected_rules: List[NodeInputValuePointerRule]
) -> None:
    rules = create_node_input_value_pointer_rules(
        descriptor,
        WorkflowDisplayContext(
            workflow_display_class=BaseWorkflowDisplay,
            workflow_display=WorkflowMetaDisplay(
                entrypoint_node_id=uuid4(),
                entrypoint_node_source_handle_id=uuid4(),
                entrypoint_node_display=NodeDisplayData(),
            ),
            global_workflow_input_displays={
                Inputs.example_workflow_input: WorkflowInputsDisplay(
                    id=UUID("a154c29d-fac0-4cd0-ba88-bc52034f5470"),
                ),
            },
            global_node_output_displays={
                MyNodeA.Outputs.output: NodeOutputDisplay(
                    id=UUID("4b16a629-11a1-4b3f-a965-a57b872d13b8"), name="output"
                ),
            },
            global_node_displays={
                MyNodeA: MyNodeADisplay(),
            },
        ),
        uuid4(),
    )
    assert rules == expected_rules


class MyState(BaseState):
    my_attribute: str


def test_create_node_input_value_pointer_rule__state_value_reference() -> None:
    """
    Tests that StateValueReference is serialized to WorkflowStatePointer using the display override ID.
    """

    # GIVEN a StateValueReference
    state_value_reference: StateValueReference[str] = MyState.my_attribute  # type: ignore[assignment]

    # AND a display context with a state value display override
    override_id = uuid4()
    display_context = WorkflowDisplayContext(
        global_state_value_displays={
            state_value_reference: StateValueDisplay(id=override_id),
        },
    )

    # WHEN we create a node input value pointer rule
    result = create_node_input_value_pointer_rule(state_value_reference, display_context)

    # THEN we should get a WorkflowStatePointer with the overridden display ID
    assert isinstance(result, WorkflowStatePointer)
    assert result.type == "WORKFLOW_STATE"
    assert isinstance(result.data, WorkflowStateData)
    assert result.data.state_variable_id == str(override_id)


class MyTrigger(BaseTrigger):
    my_attribute: str


def test_create_node_input_value_pointer_rule__trigger_attribute_reference() -> None:
    """
    Tests that TriggerAttributeReference is serialized to TriggerAttributePointer.
    """

    # GIVEN a TriggerAttributeReference
    trigger_attribute_reference: TriggerAttributeReference[str] = MyTrigger.my_attribute  # type: ignore[assignment]

    # AND a display context
    display_context = WorkflowDisplayContext()

    # WHEN we create a node input value pointer rule
    result = create_node_input_value_pointer_rule(trigger_attribute_reference, display_context)

    # THEN we should get a TriggerAttributePointer with the correct data
    assert isinstance(result, TriggerAttributePointer)
    assert result.type == "TRIGGER_ATTRIBUTE"
    assert isinstance(result.data, TriggerAttributeData)
    assert result.data.trigger_id == str(MyTrigger.__id__)
    assert result.data.attribute_id == str(trigger_attribute_reference.id)


def test_create_node_input__unexpected_error_returns_node_input_with_empty_rules() -> None:
    """
    Tests that when create_node_input_value_pointer_rules raises an unexpected error,
    create_node_input still returns a NodeInput with empty rules and adds the error to display context.
    """

    # GIVEN a display context
    display_context = WorkflowDisplayContext()

    # AND a node_id and input_name
    node_id = uuid4()
    input_name = "test_input"

    # AND create_node_input_value_pointer_rules will raise an unexpected error
    unexpected_error = RuntimeError("Unexpected error during serialization")
    with patch(
        "vellum_ee.workflows.display.nodes.vellum.utils.create_node_input_value_pointer_rules",
        side_effect=unexpected_error,
    ):
        # WHEN we call create_node_input
        result = create_node_input(
            node_id=node_id,
            input_name=input_name,
            value="test_value",
            display_context=display_context,
        )

    # THEN we should still get a NodeInput with empty rules
    assert result.key == input_name
    assert result.value.rules == []
    assert result.value.combinator == "OR"

    # AND the error should be added to the display context
    errors = list(display_context.errors)
    assert len(errors) == 1
    assert errors[0] is unexpected_error



import pytest
import sys
from uuid import uuid4

from vellum.workflows import BaseInputs, BaseNode, BaseState, BaseWorkflow, MockNodeExecution
from vellum.workflows.exceptions import NodeException
from vellum.workflows.expressions.accessor import AccessorExpression
from vellum.workflows.expressions.add import AddExpression
from vellum.workflows.expressions.and_ import AndExpression
from vellum.workflows.expressions.begins_with import BeginsWithExpression
from vellum.workflows.expressions.between import BetweenExpression
from vellum.workflows.expressions.coalesce_expression import CoalesceExpression
from vellum.workflows.expressions.concat import ConcatExpression
from vellum.workflows.expressions.contains import ContainsExpression
from vellum.workflows.expressions.does_not_begin_with import DoesNotBeginWithExpression
from vellum.workflows.expressions.does_not_contain import DoesNotContainExpression
from vellum.workflows.expressions.does_not_end_with import DoesNotEndWithExpression
from vellum.workflows.expressions.ends_with import EndsWithExpression
from vellum.workflows.expressions.in_ import InExpression
from vellum.workflows.expressions.is_blank import IsBlankExpression
from vellum.workflows.expressions.is_error import IsErrorExpression
from vellum.workflows.expressions.is_not_blank import IsNotBlankExpression
from vellum.workflows.expressions.is_not_null import IsNotNullExpression
from vellum.workflows.expressions.is_null import IsNullExpression
from vellum.workflows.expressions.length import LengthExpression
from vellum.workflows.expressions.minus import MinusExpression
from vellum.workflows.expressions.not_between import NotBetweenExpression
from vellum.workflows.expressions.not_in import NotInExpression
from vellum.workflows.expressions.or_ import OrExpression
from vellum.workflows.expressions.parse_json import ParseJsonExpression
from vellum.workflows.nodes.core.inline_subworkflow_node.node import InlineSubworkflowNode
from vellum.workflows.nodes.core.try_node.node import TryNode
from vellum.workflows.references.constant import ConstantValueReference
from vellum.workflows.references.environment_variable import EnvironmentVariableReference
from vellum.workflows.references.vellum_secret import VellumSecretReference
from vellum_ee.workflows.display.utils.expressions import base_descriptor_validator
from vellum_ee.workflows.server.virtual_file_loader import VirtualFileFinder


def test_mocks__parse_from_app__descriptors():
    # GIVEN a Base Node
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    # AND workflow inputs
    class Inputs(BaseInputs):
        bar: str

    # AND workflow state
    class State(BaseState):
        baz: str

    # AND a workflow class with that Node
    class MyWorkflow(BaseWorkflow[Inputs, State]):
        graph = StartNode

        class Outputs(BaseWorkflow.Outputs):
            final_value = StartNode.Outputs.foo

    # AND a mock workflow node execution from the app
    raw_mock_workflow_node_executions = [
        {
            "node_id": str(StartNode.__id__),
            "when_condition": {
                "type": "BINARY_EXPRESSION",
                "operator": ">=",
                "lhs": {
                    "type": "EXECUTION_COUNTER",
                    "node_id": str(StartNode.__id__),
                },
                "rhs": {
                    "type": "CONSTANT_VALUE",
                    "value": {
                        "type": "NUMBER",
                        "value": 1,
                    },
                },
            },
            "then_outputs": {
                "foo": "Hello foo",
            },
        },
        {
            "node_id": str(StartNode.__id__),
            "when_condition": {
                "type": "BINARY_EXPRESSION",
                "operator": "==",
                "lhs": {
                    "type": "WORKFLOW_INPUT",
                    "input_variable_id": str(Inputs.bar.id),
                },
                "rhs": {
                    "type": "CONSTANT_VALUE",
                    "value": {
                        "type": "STRING",
                        "value": "bar",
                    },
                },
            },
            "then_outputs": {
                "foo": "Hello bar",
            },
        },
        {
            "node_id": str(StartNode.__id__),
            "when_condition": {
                "type": "BINARY_EXPRESSION",
                "operator": "!=",
                "lhs": {
                    "type": "WORKFLOW_STATE",
                    "state_variable_id": str(State.baz.id),
                },
                "rhs": {
                    "type": "CONSTANT_VALUE",
                    "value": {
                        "type": "STRING",
                        "value": "baz",
                    },
                },
            },
            "then_outputs": {
                "foo": "Hello baz",
            },
        },
    ]

    # WHEN we parsed the raw data on `MockNodeExecution`
    node_output_mocks = MockNodeExecution.validate_all(
        raw_mock_workflow_node_executions,
        MyWorkflow,
        descriptor_validator=base_descriptor_validator,
    )

    # THEN we get the expected list of MockNodeExecution objects
    assert node_output_mocks
    assert len(node_output_mocks) == 3
    assert node_output_mocks[0] == MockNodeExecution(
        when_condition=StartNode.Execution.count.greater_than_or_equal_to(1),
        then_outputs=StartNode.Outputs(
            foo="Hello foo",
        ),
    )
    assert node_output_mocks[1] == MockNodeExecution(
        when_condition=Inputs.bar.equals("bar"),
        then_outputs=StartNode.Outputs(
            foo="Hello bar",
        ),
    )
    assert node_output_mocks[2] == MockNodeExecution(
        when_condition=State.baz.does_not_equal("baz"),
        then_outputs=StartNode.Outputs(
            foo="Hello baz",
        ),
    )


def test_mocks__parse_from_app__when_condition_defaults_to_false_without_descriptor_validator():
    """
    Tests that when_condition defaults to ConstantValueReference(False) when
    descriptor_validator is not provided. This ensures that mocks without a
    descriptor_validator have a valid when_condition that evaluates to False.
    """

    # GIVEN a Base Node
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    # AND a workflow class with that Node
    class MyWorkflow(BaseWorkflow):
        graph = StartNode

        class Outputs(BaseWorkflow.Outputs):
            final_value = StartNode.Outputs.foo

    # AND a mock workflow node execution with a valid when_condition JSON structure
    raw_mock_workflow_node_executions = [
        {
            "node_id": str(StartNode.__id__),
            "when_condition": {
                "type": "BINARY_EXPRESSION",
                "operator": ">=",
                "lhs": {
                    "type": "EXECUTION_COUNTER",
                    "node_id": str(StartNode.__id__),
                },
                "rhs": {
                    "type": "CONSTANT_VALUE",
                    "value": {
                        "type": "NUMBER",
                        "value": 1,
                    },
                },
            },
            "then_outputs": {
                "foo": "Hello foo",
            },
        },
    ]

    # WHEN we parse the raw data on `MockNodeExecution` without a descriptor_validator
    node_output_mocks = MockNodeExecution.validate_all(
        raw_mock_workflow_node_executions,
        MyWorkflow,
    )

    # THEN we get a list of MockNodeExecution objects
    assert node_output_mocks is not None
    assert len(node_output_mocks) == 1

    # AND the when_condition defaults to ConstantValueReference(False)
    assert node_output_mocks[0].when_condition == ConstantValueReference(False)


def test_mocks__use_node_id_from_display():
    """
    Tests that validate_all correctly resolves mocks when the node ID is annotated by the display class.
    """

    # GIVEN a workflow module with a display class that sets a custom node_id
    display_node_id = uuid4()
    files = {
        "__init__.py": "",
        "workflow.py": """\
from vellum.workflows import BaseWorkflow
from vellum.workflows.nodes import BaseNode


class StartNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        foo: str


class Workflow(BaseWorkflow):
    graph = StartNode

    class Outputs(BaseWorkflow.Outputs):
        final_value = StartNode.Outputs.foo
""",
        "display/__init__.py": """\
# flake8: noqa: F401, F403

from .workflow import *
from .nodes import *
""",
        "display/workflow.py": """\
from vellum_ee.workflows.display.workflows import BaseWorkflowDisplay
from ..workflow import Workflow


class WorkflowDisplay(BaseWorkflowDisplay[Workflow]):
    pass
""",
        "display/nodes/__init__.py": """\
# flake8: noqa: F401, F403

from .start_node import *
""",
        "display/nodes/start_node.py": f"""\
from uuid import UUID
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay
from ...workflow import StartNode


class StartNodeDisplay(BaseNodeDisplay[StartNode]):
    node_id = UUID("{display_node_id}")
""",
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered
    sys.meta_path.append(VirtualFileFinder(files, namespace))

    # AND the workflow is loaded from the module
    Workflow = BaseWorkflow.load_from_module(namespace)
    StartNode = list(Workflow.get_nodes())[0]

    # AND a mock workflow node execution using the display-annotated node ID
    raw_mock_workflow_node_executions = [
        {
            "node_id": str(display_node_id),
            "when_condition": {
                "type": "BINARY_EXPRESSION",
                "operator": ">=",
                "lhs": {
                    "type": "EXECUTION_COUNTER",
                    "node_id": str(display_node_id),
                },
                "rhs": {
                    "type": "CONSTANT_VALUE",
                    "value": {
                        "type": "NUMBER",
                        "value": 0,
                    },
                },
            },
            "then_outputs": {
                "foo": "Hello",
            },
        },
    ]

    # WHEN we parse the mock workflow node execution
    node_output_mocks = MockNodeExecution.validate_all(
        raw_mock_workflow_node_executions,
        Workflow,
        descriptor_validator=base_descriptor_validator,
    )

    # THEN we get the expected list of MockNodeExecution objects
    assert node_output_mocks
    assert len(node_output_mocks) == 1
    assert node_output_mocks[0] == MockNodeExecution(
        when_condition=StartNode.Execution.count.greater_than_or_equal_to(0),
        then_outputs=StartNode.Outputs(  # type: ignore[call-arg]
            foo="Hello",
        ),
    )


def test_mocks__node_not_found_in_workflow_skips_with_warning(caplog):
    """
    Tests that when a mock references a node_id that doesn't exist in the workflow,
    the mock is skipped with a warning instead of raising an error.
    """

    # GIVEN a Base Node
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    # AND a workflow class with that Node
    class MyWorkflow(BaseWorkflow):
        graph = StartNode

        class Outputs(BaseWorkflow.Outputs):
            final_value = StartNode.Outputs.foo

    # AND a mock workflow node execution referencing a non-existent node_id
    non_existent_node_id = uuid4()
    raw_mock_with_missing_node = [
        {
            "node_id": str(non_existent_node_id),
            "when_condition": {
                "type": "BINARY_EXPRESSION",
                "operator": ">=",
                "lhs": {
                    "type": "EXECUTION_COUNTER",
                    "node_id": str(non_existent_node_id),
                },
                "rhs": {
                    "type": "CONSTANT_VALUE",
                    "value": {
                        "type": "NUMBER",
                        "value": 0,
                    },
                },
            },
            "then_outputs": {
                "foo": "Hello",
            },
        }
    ]

    # WHEN we parse the mock workflow node execution
    node_output_mocks = MockNodeExecution.validate_all(
        raw_mock_with_missing_node,
        MyWorkflow,
        descriptor_validator=base_descriptor_validator,
    )

    # THEN we get an empty list (the mock was skipped)
    assert node_output_mocks == []

    # AND a warning was logged
    assert any(
        f"Skipping mock for node {non_existent_node_id}" in record.message
        and "node not found in workflow MyWorkflow" in record.message
        for record in caplog.records
    )


def test_base_descriptor_validator__vellum_secret():
    """
    Tests that VELLUM_SECRET descriptor type is correctly validated.
    """

    # GIVEN a simple workflow
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    class MyWorkflow(BaseWorkflow):
        graph = StartNode

    # AND a VELLUM_SECRET descriptor
    raw_descriptor = {
        "type": "VELLUM_SECRET",
        "vellum_secret_name": "my_secret",
    }

    # WHEN we validate the descriptor
    result = base_descriptor_validator(raw_descriptor, MyWorkflow)

    # THEN we get a VellumSecretReference
    assert isinstance(result, VellumSecretReference)
    assert result.name == "my_secret"


def test_base_descriptor_validator__environment_variable():
    """
    Tests that ENVIRONMENT_VARIABLE descriptor type is correctly validated.
    """

    # GIVEN a simple workflow
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    class MyWorkflow(BaseWorkflow):
        graph = StartNode

    # AND an ENVIRONMENT_VARIABLE descriptor
    raw_descriptor = {
        "type": "ENVIRONMENT_VARIABLE",
        "environment_variable": "MY_ENV_VAR",
    }

    # WHEN we validate the descriptor
    result = base_descriptor_validator(raw_descriptor, MyWorkflow)

    # THEN we get an EnvironmentVariableReference
    assert isinstance(result, EnvironmentVariableReference)
    assert result.name == "MY_ENV_VAR"


def test_base_descriptor_validator__array_reference():
    """
    Tests that ARRAY_REFERENCE descriptor type is correctly validated.
    """

    # GIVEN a simple workflow
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    class MyWorkflow(BaseWorkflow):
        graph = StartNode

    # AND an ARRAY_REFERENCE descriptor with constant values
    raw_descriptor = {
        "type": "ARRAY_REFERENCE",
        "items": [
            {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "item1"}},
            {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "item2"}},
        ],
    }

    # WHEN we validate the descriptor
    result = base_descriptor_validator(raw_descriptor, MyWorkflow)

    # THEN we get a ConstantValueReference containing a list
    assert isinstance(result, ConstantValueReference)
    assert isinstance(result._value, list)
    assert len(result._value) == 2


def test_base_descriptor_validator__dictionary_reference():
    """
    Tests that DICTIONARY_REFERENCE descriptor type is correctly validated.
    """

    # GIVEN a simple workflow
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    class MyWorkflow(BaseWorkflow):
        graph = StartNode

    # AND a DICTIONARY_REFERENCE descriptor
    raw_descriptor = {
        "type": "DICTIONARY_REFERENCE",
        "entries": [
            {
                "id": "entry1",
                "key": "key1",
                "value": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "value1"}},
            },
            {
                "id": "entry2",
                "key": "key2",
                "value": {"type": "CONSTANT_VALUE", "value": {"type": "NUMBER", "value": 42}},
            },
        ],
    }

    # WHEN we validate the descriptor
    result = base_descriptor_validator(raw_descriptor, MyWorkflow)

    # THEN we get a ConstantValueReference containing a dict
    assert isinstance(result, ConstantValueReference)
    assert isinstance(result._value, dict)
    assert "key1" in result._value
    assert "key2" in result._value


@pytest.mark.parametrize(
    "operator,expected_type",
    [
        ("blank", IsBlankExpression),
        ("notBlank", IsNotBlankExpression),
        ("null", IsNullExpression),
        ("notNull", IsNotNullExpression),
        ("isError", IsErrorExpression),
        ("length", LengthExpression),
        ("parseJson", ParseJsonExpression),
    ],
)
def test_base_descriptor_validator__unary_expression(operator, expected_type):
    """
    Tests that UNARY_EXPRESSION descriptor types are correctly validated.
    """

    # GIVEN a simple workflow
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    class MyWorkflow(BaseWorkflow):
        graph = StartNode

    # AND a UNARY_EXPRESSION descriptor
    raw_descriptor = {
        "type": "UNARY_EXPRESSION",
        "operator": operator,
        "lhs": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "test"}},
    }

    # WHEN we validate the descriptor
    result = base_descriptor_validator(raw_descriptor, MyWorkflow)

    # THEN we get the expected expression type
    assert isinstance(result, expected_type)


@pytest.mark.parametrize(
    "operator,expected_type",
    [
        ("between", BetweenExpression),
        ("notBetween", NotBetweenExpression),
    ],
)
def test_base_descriptor_validator__ternary_expression(operator, expected_type):
    """
    Tests that TERNARY_EXPRESSION descriptor types are correctly validated.
    """

    # GIVEN a simple workflow
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    class MyWorkflow(BaseWorkflow):
        graph = StartNode

    # AND a TERNARY_EXPRESSION descriptor
    raw_descriptor = {
        "type": "TERNARY_EXPRESSION",
        "operator": operator,
        "base": {"type": "CONSTANT_VALUE", "value": {"type": "NUMBER", "value": 5}},
        "lhs": {"type": "CONSTANT_VALUE", "value": {"type": "NUMBER", "value": 1}},
        "rhs": {"type": "CONSTANT_VALUE", "value": {"type": "NUMBER", "value": 10}},
    }

    # WHEN we validate the descriptor
    result = base_descriptor_validator(raw_descriptor, MyWorkflow)

    # THEN we get the expected expression type
    assert isinstance(result, expected_type)


@pytest.mark.parametrize(
    "operator,expected_type",
    [
        ("=", BeginsWithExpression.__bases__[0].__bases__[0]),
        ("==", BeginsWithExpression.__bases__[0].__bases__[0]),
        ("doesNotContain", DoesNotContainExpression),
        ("doesNotBeginWith", DoesNotBeginWithExpression),
        ("doesNotEndWith", DoesNotEndWithExpression),
        ("in", InExpression),
        ("notIn", NotInExpression),
        ("and", AndExpression),
        ("or", OrExpression),
        ("coalesce", CoalesceExpression),
        ("+", AddExpression),
        ("-", MinusExpression),
        ("concat", ConcatExpression),
    ],
)
def test_base_descriptor_validator__binary_expression_additional_operators(operator, expected_type):
    """
    Tests that additional BINARY_EXPRESSION operators are correctly validated.
    """

    # GIVEN a simple workflow
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    class MyWorkflow(BaseWorkflow):
        graph = StartNode

    # AND a BINARY_EXPRESSION descriptor with the given operator
    raw_descriptor = {
        "type": "BINARY_EXPRESSION",
        "operator": operator,
        "lhs": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "left"}},
        "rhs": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "right"}},
    }

    # WHEN we validate the descriptor
    result = base_descriptor_validator(raw_descriptor, MyWorkflow)

    # THEN we get the expected expression type
    assert isinstance(result, expected_type)


def test_base_descriptor_validator__accessor_expression():
    """
    Tests that accessField operator in BINARY_EXPRESSION is correctly validated.
    """

    # GIVEN a simple workflow
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    class MyWorkflow(BaseWorkflow):
        graph = StartNode

    # AND a BINARY_EXPRESSION descriptor with accessField operator
    raw_descriptor = {
        "type": "BINARY_EXPRESSION",
        "operator": "accessField",
        "lhs": {"type": "CONSTANT_VALUE", "value": {"type": "JSON", "value": {"field": "value"}}},
        "rhs": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "field"}},
    }

    # WHEN we validate the descriptor
    result = base_descriptor_validator(raw_descriptor, MyWorkflow)

    # THEN we get an AccessorExpression
    assert isinstance(result, AccessorExpression)


def test_mocks__validate_all__node_nested_in_subworkflow():
    """
    Tests that MockNodeExecution.validate_all correctly handles mocks for nodes
    nested within a subworkflow node by validating against the inner subworkflow.
    """

    # GIVEN a node that will be nested inside a subworkflow
    class NestedNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            result: str

        def run(self) -> Outputs:
            raise NodeException("This node should be mocked")

    # AND a subworkflow containing that nested node
    class InnerSubworkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = NestedNode

        class Outputs(BaseWorkflow.Outputs):
            inner_result = NestedNode.Outputs.result

    # AND a subworkflow node that uses the inner subworkflow
    class SubworkflowNode(InlineSubworkflowNode):
        subworkflow = InnerSubworkflow

    # AND an outer workflow containing the subworkflow node
    class OuterWorkflow(BaseWorkflow):
        graph = SubworkflowNode

        class Outputs(BaseWorkflow.Outputs):
            final_result = SubworkflowNode.Outputs.inner_result

    # AND raw mock data for the nested node using the modern data model format
    raw_mock_workflow_node_executions = [
        {
            "node_id": str(NestedNode.__id__),
            "when_condition": {
                "type": "BINARY_EXPRESSION",
                "operator": ">=",
                "lhs": {
                    "type": "EXECUTION_COUNTER",
                    "node_id": str(NestedNode.__id__),
                },
                "rhs": {
                    "type": "CONSTANT_VALUE",
                    "value": {
                        "type": "NUMBER",
                        "value": 0,
                    },
                },
            },
            "then_outputs": {
                "result": "mocked_result",
            },
        }
    ]

    # WHEN we call validate_all on the outer workflow
    node_output_mocks = MockNodeExecution.validate_all(
        raw_mock_workflow_node_executions,
        OuterWorkflow,
        descriptor_validator=base_descriptor_validator,
    )

    # THEN we get a list of MockNodeExecution objects
    assert node_output_mocks is not None
    assert len(node_output_mocks) == 1

    # AND the mock is correctly parsed with the nested node's outputs
    assert node_output_mocks[0] == MockNodeExecution(
        when_condition=NestedNode.Execution.count.greater_than_or_equal_to(0),
        then_outputs=NestedNode.Outputs(result="mocked_result"),
    )

    # AND when we run the outer workflow with the mocks
    workflow = OuterWorkflow()
    terminal_event = workflow.run(node_output_mocks=node_output_mocks)

    # THEN the workflow completes successfully
    assert terminal_event.name == "workflow.execution.fulfilled", terminal_event

    # AND the output reflects the mocked value from the nested node
    assert terminal_event.outputs.final_result == "mocked_result"


def test_mocks__serialize__try_node_wrapped_node_has_correct_node_id():
    """
    Tests that when serializing a MockNodeExecution for a node wrapped in a TryNode adornment,
    the serialized node_id matches the inner wrapped node's ID (not the adornment wrapper's ID).
    """

    # GIVEN a node wrapped in a TryNode adornment
    @TryNode.wrap()
    class WrappedNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            result: str

    # AND a workflow that uses the wrapped node
    class MyWorkflow(BaseWorkflow):
        graph = WrappedNode

        class Outputs(BaseWorkflow.Outputs):
            final_result = WrappedNode.Outputs.result

    # AND a mock for the wrapped node
    mock = MockNodeExecution(
        when_condition=ConstantValueReference(True),
        then_outputs=WrappedNode.Outputs(result="mocked_result"),
    )

    # WHEN we serialize the mock
    serialized_mock = mock.model_dump()

    # THEN the serialized node_id should match the inner wrapped node's ID
    inner_node = WrappedNode.__wrapped_node__
    assert inner_node is not None
    assert serialized_mock["node_id"] == str(inner_node.__id__)

    # AND the serialized mock should have the correct type
    assert serialized_mock["type"] == "NODE_EXECUTION"

    # AND the then_outputs should be serialized correctly
    assert serialized_mock["then_outputs"]["result"] == "mocked_result"


def test_mocks__validate_all__try_node_wrapped_node_deserializes_correctly():
    """
    Tests that validate_all correctly deserializes a mock for a node wrapped in a TryNode adornment.
    """

    # GIVEN a node wrapped in a TryNode adornment
    @TryNode.wrap()
    class WrappedNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            result: str

    # AND a workflow that uses the wrapped node
    class MyWorkflow(BaseWorkflow):
        graph = WrappedNode

        class Outputs(BaseWorkflow.Outputs):
            final_result = WrappedNode.Outputs.result

    # AND the inner wrapped node's ID
    inner_node = WrappedNode.__wrapped_node__
    assert inner_node is not None

    # AND a raw mock workflow node execution using the inner node's ID
    raw_mock_workflow_node_executions = [
        {
            "node_id": str(inner_node.__id__),
            "when_condition": {
                "type": "BINARY_EXPRESSION",
                "operator": ">=",
                "lhs": {
                    "type": "EXECUTION_COUNTER",
                    "node_id": str(inner_node.__id__),
                },
                "rhs": {
                    "type": "CONSTANT_VALUE",
                    "value": {
                        "type": "NUMBER",
                        "value": 1,
                    },
                },
            },
            "then_outputs": {
                "result": "mocked_result",
            },
        },
    ]

    # WHEN we parse the raw data on MockNodeExecution
    node_output_mocks = MockNodeExecution.validate_all(
        raw_mock_workflow_node_executions,
        MyWorkflow,
        descriptor_validator=base_descriptor_validator,
    )

    # THEN we get a list with one MockNodeExecution object
    assert node_output_mocks is not None
    assert len(node_output_mocks) == 1

    # AND the MockNodeExecution has the correct when_condition
    assert node_output_mocks[0].when_condition == inner_node.Execution.count.greater_than_or_equal_to(1)

    # AND the then_outputs is the correct type with the expected value
    assert node_output_mocks[0].then_outputs == inner_node.Outputs(result="mocked_result")


def test_mocks__validate_all__ignores_undefined_outputs():
    """
    Tests that validate_all ignores outputs that are not defined in the node's Outputs class.
    """

    # GIVEN a node wrapped in a TryNode adornment
    @TryNode.wrap()
    class WrappedNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            result: str

    # AND a workflow that uses the wrapped node
    class MyWorkflow(BaseWorkflow):
        graph = WrappedNode

        class Outputs(BaseWorkflow.Outputs):
            final_result = WrappedNode.Outputs.result

    # AND the inner wrapped node's ID
    inner_node = WrappedNode.__wrapped_node__
    assert inner_node is not None

    # AND a raw mock workflow node execution with an undefined "error" output
    raw_mock_workflow_node_executions = [
        {
            "node_id": str(inner_node.__id__),
            "when_condition": {
                "type": "BINARY_EXPRESSION",
                "operator": ">=",
                "lhs": {
                    "type": "EXECUTION_COUNTER",
                    "node_id": str(inner_node.__id__),
                },
                "rhs": {
                    "type": "CONSTANT_VALUE",
                    "value": {
                        "type": "NUMBER",
                        "value": 1,
                    },
                },
            },
            "then_outputs": {
                "result": "mocked_result",
                "error": "some error value",
            },
        },
    ]

    # WHEN we parse the raw data on MockNodeExecution
    node_output_mocks = MockNodeExecution.validate_all(
        raw_mock_workflow_node_executions,
        MyWorkflow,
        descriptor_validator=base_descriptor_validator,
    )

    # THEN we get a list with one MockNodeExecution object
    assert node_output_mocks is not None
    assert len(node_output_mocks) == 1

    # AND the then_outputs only contains the declared "result" output
    assert node_output_mocks[0].then_outputs == inner_node.Outputs(result="mocked_result")

    # AND the "error" output was ignored and is not present
    assert not hasattr(node_output_mocks[0].then_outputs, "error")
